# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

from qsteed.dag.circuit_dag_convert import circuit_to_dag, dag_to_circuit
from qsteed.dag.circuit_dag_convert import node_to_gate, gate_to_node
from qsteed.dag.dagcircuit import DAGCircuit
from qsteed.passes.basepass import BasePass
from qsteed.passes.decomposition.ZYZ_decompose import zyz_decomposition, xyx_decomposition, zxz_decomposition
from quafu import QuantumCircuit
from quafu.elements.element_gates.rotation import RZGate, RYGate, RXGate


class OneQubitGateOptimization(BasePass):
    """
    Remove redundant gates
    """

    def run(self, circuit: QuantumCircuit) -> QuantumCircuit:
        """
        Transform the circuit into a DAG, remove redundant nodes, and transform the DAG back to a circuit.

        Args:
            circuit (QuantumCircuit): The input quantum circuit.

        Returns:
            QuantumCircuit: The optimized quantum circuit with redundant gates removed.
        """
        # new_circuit = circuit_to_dag(circuit, measure_flag=True)
        # new_circuit = self._remove_one_qubit_Gate(circuit=new_circuit)
        # new_circuit = dag_to_circuit(new_circuit, circuit.num)

        dag = circuit_to_dag(circuit, measure_flag=True)
        optimized_dag = self._remove_one_qubit_Gate(dag)
        optimized_circuit = dag_to_circuit(optimized_dag, circuit.num)
        return optimized_circuit

    def _sorted_node_list(self, circuit: DAGCircuit) -> dict:
        """
        Sort and return the nodes of the DAGCircuit.

        Args:
            circuit (DAGCircuit): The input DAGCircuit.

        Returns:
            dict: A sorted dictionary of nodes from the DAGCircuit.
        """
        node_list = circuit.nodes_dict()
        if 'm' in node_list:
            sorted_node = {k: node_list[k] for k in list(node_list.keys())[:-1]}
            sorted_node = {k: sorted_node[k] for k in sorted(sorted_node)}
        else:
            sorted_node = {k: node_list[k] for k in sorted(node_list)}
        return sorted_node

    # def _sorted_node_list(self, circuit: DAGCircuit) -> dict:
    #     """
    #     Sort and return the nodes of the DAGCircuit.
    #
    #     Args:
    #         circuit (DAGCircuit): The input DAGCircuit.
    #
    #     Returns:
    #         dict: A sorted dictionary of nodes from the DAGCircuit.
    #     """
    #     node_list = circuit.nodes_dict()
    #     sorted_keys = sorted(node_list.keys())
    #
    #     if 'm' in node_list:
    #         sorted_keys.remove('m')
    #         sorted_keys.append('m')
    #
    #     sorted_node = {k: node_list[k] for k in sorted_keys}
    #     return sorted_node

    '''
    This function is to combine continous one_qubit_gate
    '''

    def _remove_one_qubit_Gate(self, circuit: DAGCircuit) -> DAGCircuit:
        sorted_node = self._sorted_node_list(circuit)
        one_qubit_gate = ['H', 'RX', 'RZ', 'RY']
        gate_record = {}
        # loop all the nodes
        for item in sorted_node.items():
            node = item[1]
            # add one qubit gate
            if node.name in one_qubit_gate:
                successors = circuit.node_qubits_successors(node)
                # check current qubit has previous one-qubit gate
                pos = node.pos[0]
                gate = node_to_gate(copy.deepcopy(node))
                if pos in gate_record and gate_record[pos]:
                    # for each item in gate_record, the format is
                    # {key: [successors, num_combined_gate, combined_matrix, nodes]}
                    gate_record[pos][-1].append(node)
                    gate_record[pos] = [successors, gate_record[pos][1] + 1, gate.matrix @ gate_record[pos][2],
                                        gate_record[pos][-1]]
                else:
                    gate_record[pos] = [successors, 1, gate.matrix, [node]]
            else:
                # for this qubit, encounter multi-qubit gate, now just add the combined gate back to
                for pos in node.pos:
                    if pos in gate_record and gate_record[pos] and len(gate_record[pos][-1]) > 3:
                        # to find the minimum gate based on the angle, for RX, RY gate,
                        # when the theta is 0, it is the same as indentity matrix
                        self.gate_check(circuit=circuit, gate_record=gate_record, pos=pos)
                    gate_record[pos] = None
        # at the end of list, clean up record in case some gates are missing 
        for pos in gate_record.keys():
            if gate_record[pos] and len(gate_record[pos][-1]) > 3:
                self.gate_check(circuit=circuit, gate_record=gate_record, pos=pos)
            gate_record[pos] = []
        return circuit

    def gate_check(self, circuit, gate_record, pos, node=None):
        '''
            use zyz, zxz, xyx to decompose the gates and check which function will provide the minimum number of gates
            only RX, RY will be removed when the angle is 0. For zyz and zxz, if the middle angle is zero, we only have one gate.
        '''
        node = gate_record[pos][3][0]
        predecessors = circuit.node_qubits_predecessors(node)

        label = []
        for gate in gate_record[pos][-1]:
            label.append(gate.label)
            circuit.remove_instruction_node(gate)

        # gamma, beta, alpha, global_phase = xyx_decomposition(gate_record[pos][2])
        # if gamma * beta * alpha == 0:
        #     if gamma != 0:
        #         circuit.add_instruction_node(gate_to_node(RXGate(pos, gamma), label[0]),
        #                                      predecessors, gate_record[pos][0])
        #         label = label[1:]
        #     if beta != 0:
        #         circuit.add_instruction_node(gate_to_node(RYGate(pos, beta), label[0]),
        #                                      predecessors, gate_record[pos][0])
        #         label = label[1:]
        #     if alpha != 0:
        #         circuit.add_instruction_node(gate_to_node(RXGate(pos, alpha), label[0]),
        #                                      predecessors, gate_record[pos][0])
        #     return

        gamma, beta, alpha, global_phase = xyx_decomposition(gate_record[pos][2])
        if beta == 0:
            circuit.add_instruction_node(gate_to_node(RXGate(pos, gamma + alpha), label[0]),
                                         predecessors, gate_record[pos][0])
            return

        gamma, beta, alpha, global_phase = zxz_decomposition(gate_record[pos][2])
        if beta == 0:
            circuit.add_instruction_node(gate_to_node(RZGate(pos, gamma + alpha), label[0]),
                                         predecessors, gate_record[pos][0])
            return

        gamma, beta, alpha, global_phase = zyz_decomposition(gate_record[pos][2])
        if beta == 0:
            circuit.add_instruction_node(gate_to_node(RZGate(pos, gamma + alpha), label[0]),
                                         predecessors, gate_record[pos][0])
            return

        else:
            node1 = gate_to_node(RZGate(pos, gamma), label[0])
            circuit.add_instruction_node(node1, predecessors, gate_record[pos][0])
            node2 = gate_to_node(RYGate(pos, beta), label[1])
            circuit.add_instruction_node(node2, {pos: node1}, gate_record[pos][0])
            node3 = gate_to_node(RZGate(pos, alpha), label[2])
            circuit.add_instruction_node(node3, {pos: node2}, gate_record[pos][0])
        return



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from math import pi

from quafu import QuantumCircuit
from quafu.elements import Barrier, Delay, XYResonance, Measure
from quafu.exceptions.quafu_error import CompileError

from qsteed.passes.basepass import BasePass
from qsteed.passes.unroll.rules_library import Rules_dict


class UnrollTo2Qubit(BasePass):
    """Unroll a circuit to two-qubit gates.
    """

    def __init__(self):
        """Initialize UnrollTo2Qubit
        """
        super().__init__()
        self.quantum_element = [Barrier.name, Delay.name, XYResonance.name, Measure.name]
        self.gate_run_limit = 8
        self.qubits = None
        self.global_phase = 0

    def run(self, circuit: QuantumCircuit) -> QuantumCircuit:
        gates = circuit.gates
        self.qubits = circuit.num
        new_circuit = QuantumCircuit(circuit.num)
        for op in gates:
            if (op.name.lower() in self.quantum_element) or (get_length(op.pos) <= 2):
                new_circuit.add_gate(op)
            else:
                current_op = copy.deepcopy(op)
                new_circuit = self._apply_gate_rules(current_op, op, new_circuit=new_circuit, depth=self.gate_run_limit)
        [new_circuit.measure([q], [c]) for q, c in circuit.measures.items()]
        self.global_phase = self.global_phase % (2 * pi)
        return new_circuit

    def _apply_gate_rules(self, current_op, gate, new_circuit=None, depth=5):
        if new_circuit is None:
            new_circuit = QuantumCircuit(self.qubits)

        if depth <= 0:
            # Exceeding the number of recursions indicates that no decomposition of this gate
            # based on 1-qubit and 2-qubit gates set has been found among all rules.
            raise CompileError("The instruction %s cannot be unrolled to 2-qubit gates." % (
                current_op.name.lower()))

        if (gate.name.lower() in self.quantum_element) or (get_length(gate.pos) <= 2):
            new_circuit.add_gate(gate)
            return new_circuit
        elif gate.name.lower() == 'unitary':
            # TODO: Currently quafu has no definition of unitary gate.
            pass
        else:
            try:
                Rules_dict[gate.name.lower()][0]
            except KeyError:
                print("Error: The instruction %s has no unrolling rule." % gate.name)
            rule_class = Rules_dict[gate.name.lower()][1]
            rule = rule_class.run(gate)
            self.global_phase += rule_class.global_phase
            for gate in rule:
                new_circuit = self._apply_gate_rules(current_op, gate, new_circuit, depth=depth - 1)
            return new_circuit


def get_length(obj):
    if isinstance(obj, list) or isinstance(obj, dict):
        return len(obj)
    elif isinstance(obj, int):
        return 1
    else:
        return None



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the TranspilerVis class"""

from qsteed.transpiler.transpiler_visualization import TranspilerVis, dynamic_draw, draw_allpass_circuits
from .shared_utils import get_random_circuit, get_passflow, get_initial_model


class TestTranspilerVis:
    """Test cases for the TranspilerVis class and related visualization functionalities."""
    def test_transpiler_vis(self):
        """Test the transpilation visualization process."""
        qc = get_random_circuit()
        passflow = get_passflow()
        initial_model = get_initial_model()

        transpiler_vis = TranspilerVis(passflow, initial_model)
        transpiled_circuit, info, short_info = transpiler_vis.transpile_vis(qc)

        assert transpiled_circuit is not None

        print("The information about the input index pass, here the input index is dynamic")
        # dynamic_draw: Display intermediate compilation results interactively. Please uncomment when in use.
        # dynamic_draw(info, short_info)
        draw_allpass_circuits(info)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
import numpy as np
from quafu import QuantumCircuit
from quafu.elements import Barrier, Measure


def relabel_graph(graph):
    """Standardize the graph using degrees as node weights.
    """
    for nodes in graph.nodes.data():
        graph.add_node(nodes[0], weight=graph.degree[nodes[0]])
    return graph


def circuit_to_graph(circuit: QuantumCircuit):
    """
    Convert the circuit to a weighted graph, with the edge weights
     being the corresponding number of two-qubit gates.
     The multi-qubit gate becomes a weighted graph,
     which is equivalent to the full connection of each qubit.
    """
    num_qubits = circuit.num
    temp = np.zeros((num_qubits, num_qubits))
    gates_list = circuit.gates
    for gate in gates_list:
        if gate.name not in [Barrier.name, Measure.name]:
            if isinstance(gate.pos, list) and len(gate.pos) >= 2:
                index = sorted(gate.pos)
                for i in range(len(index)):
                    for j in range(i + 1, len(index)):
                        temp[index[i]][index[j]] = temp[index[i]][index[j]] + 1
            # else:
            #     index = gate.pos
            #     temp[index][index] = temp[index][index] + 1
    graph = nx.Graph(temp)
    graph = relabel_graph(graph)
    return graph


def draw_graph(g):
    """Draw weighted graph
    """
    import matplotlib.pyplot as plt
    label_name = list(list(g.edges(data=True))[0][2].keys())[0]
    labels = nx.get_edge_attributes(g, label_name)
    pos = nx.spring_layout(g)
    # pos = nx.circular_layout(g)
    nx.draw_networkx(g, pos=pos, with_labels=True, node_color='r', node_size=200, edge_color='b',
                     width=1, font_size=10)
    if isinstance(g, nx.MultiDiGraph) or isinstance(g, nx.MultiGraph):
        pass
    else:
        nx.draw_networkx_edge_labels(g, pos=pos, edge_labels=labels)
    plt.show()



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import pi
from typing import List

import numpy as np
from quafu.elements import Instruction
from quafu.elements.element_gates import MCRXGate, CRXGate, HGate, ToffoliGate, CPGate, RYGate, SGate, SdgGate, CXGate

from qsteed.passes.basepass import UnrollPass


class MCRXToCNOT(UnrollPass):
    """
    The MCRXtoCNOT pass.
    convert MCRX gate to {CNOT, H, Rz}.

    MCRX gate decomposition rule :
    (control bit 0:1, target bit: 3)
    q[0]  --*--          q[0]  ------------------------*---------------------*----*-----T-----*---------------------------*---------------------*----*----T------*------
            |                                          |                     |    |           |                           |                     |    |           |
    q[1]  --*--  ≡       q[1]  ------------*-----------|---------*-----T-----|----+----Tdg----+---------------*-----------|---------*----T------|----+----Tdg----+------
            |                              |           |         |           |                                |           |         |           |
    q[2]  --RX(θ)--      q[2]  --S----H----+----Tdg----+----T----+----Tdg----+----T-----H----RY(-θ/2)----H----+----Tdg----+----T----+----Tdg----+----T----H----RY(θ/2)--

    (control bit 0:2, target bit: 3)
    q[0]  --*--          q[0]  --------------------------------------+----*-----+----*----+-----*----+----*-----------------------------------------------------+----*-----+----*----+-----*----+----*--------------------------
            |                                                        |    |     |    |    |     |    |    |                                                     |    |     |    |    |     |    |    |
    q[1]  --*--  ≡       q[1]  ------------------+----*----+----*----*----|-----|----|----*-----|----|----|---------------------------------+----*----+----*----*----|-----|----|----*-----|----|----|--------------------------
            |                                    |    |    |    |         |     |    |          |    |    |                                 |    |    |    |         |     |    |          |    |    |
    q[2]  --*--          q[2]  ------------*-----*----|----*----|---------|-----*----|----------|----*----|---------------------------*-----*----|----*----|---------|-----*----|----------|----*----|--------------------------
            |                              |          |         |         |          |          |         |                           |          |         |         |          |          |         |
    q[3]  --RX(θ)--      q[3]  --S----H----R1--------R1_--------R1--------R1_--------R1--------R1_--------R1----H----RY(-θ/2)----H----R1--------R1_--------R1--------R1_--------R1--------R1_--------R1----H----RY(θ/2)----Sdg--
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = MCRXGate([0, 1], 2, 0).name.lower()
        self.basis = [CXGate.name.lower(), HGate.name.lower(), CPGate(0, 1, 0).name.lower(),
                      ToffoliGate(0, 1, 2).name.lower(), RYGate(0, 0).name.lower(), SGate.name.lower(),
                      SdgGate.name.lower()]

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, MCRXGate):
            control_bits = op.ctrls
            if isinstance(op.paras, list):
                theta = op.paras[0]
            else:
                theta = op.paras
            if isinstance(op.targs, list):
                target_bit = op.targs[0]
            else:
                target_bit = op.targs

            if len(control_bits) == 1:
                rule.append(CRXGate(control_bits[0], target_bit))
            elif len(control_bits) == 2:
                rule.append(SGate(target_bit))
                rule.append(ToffoliGate(control_bits[0], control_bits[1], target_bit))
                rule.append(RYGate(target_bit, -theta / 2))
                rule.append(ToffoliGate(control_bits[0], control_bits[1], target_bit))
                rule.append(RYGate(target_bit, theta / 2))
                rule.append(SdgGate(target_bit))
            else:
                graycode = self.build_gray_code(len(control_bits))
                theta = pi / 2 ** (len(control_bits) - 1)
                last_code = graycode[0, :]
                rule.append(SGate(target_bit))
                rule.append(HGate(target_bit))
                # loop the gray code based on the line order
                for i in range(1, len(graycode)):
                    code = graycode[i, :]
                    ones_bits = np.where(code == 1)[0]
                    set_idx = ones_bits[0]
                    diff_idx = np.where(code != last_code)[0][0]
                    if diff_idx != set_idx:
                        rule.append(CXGate(control_bits[diff_idx], control_bits[set_idx]))
                    elif len(ones_bits) >= 2:
                        next_idx = ones_bits[1]
                        rule.append(CXGate(control_bits[next_idx], control_bits[set_idx]))
                    if np.sum(code) % 2 == 0:
                        rule.append(CPGate(control_bits[set_idx], target_bit, -theta))
                    else:
                        rule.append(CPGate(control_bits[set_idx], target_bit, theta))
                    last_code = code
                rule.append(HGate(target_bit))
                rule.append(RYGate(target_bit, -theta / 2))
                rule.append(HGate(target_bit))
                last_code = graycode[0, :]
                for i in range(1, len(graycode)):
                    code = graycode[i, :]
                    ones_bits = np.where(code == 1)[0]
                    set_idx = ones_bits[0]
                    diff_idx = np.where(code != last_code)[0][0]
                    if diff_idx != set_idx:
                        rule.append(CXGate(control_bits[diff_idx], control_bits[set_idx]))
                    elif len(ones_bits) >= 2:
                        next_idx = ones_bits[1]
                        rule.append(CXGate(control_bits[next_idx], control_bits[set_idx]))
                    if np.sum(code) % 2 == 0:
                        rule.append(CPGate(control_bits[set_idx], target_bit, -theta))
                    else:
                        rule.append(CPGate(control_bits[set_idx], target_bit, theta))
                    last_code = code
                rule.append(HGate(target_bit))
                rule.append(RYGate(target_bit, theta / 2))
                rule.append(SdgGate(target_bit))
        else:
            rule.append(op)
        self.rule = rule
        return rule

    # generate the gray code to assign control gate between control bits
    def build_gray_code(self, n):
        code = np.full((2 ** n, n), 0)
        block = np.repeat([0, 1], 2 ** (n - 1))
        code[:, 0] = block
        for i in range(1, n):
            block = np.repeat([0, 1, 1, 0], 2 ** (n - 1 - i))
            print(block)
            code[:, i] = np.tile(block, 2 ** (i - 1))
        return code



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from quafu import QuantumCircuit

from qsteed.dag.circuit_dag_convert import circuit_to_dag, dag_to_circuit
from qsteed.dag.dagcircuit import DAGCircuit
from qsteed.passes.basepass import BasePass


class GateCombineOptimization(BasePass):
    """
    Remove redundant gates
    """

    '''
    Here we only consider the basic gate since the previous step has transformed the gates into basic gates
    '''

    def run(self, circuit: QuantumCircuit) -> QuantumCircuit:
        """
        Transform the circuit into dag, remove redandunt node, then transform the dag back to circuit
        """
        new_circuit = circuit_to_dag(circuit, measure_flag=True)
        new_circuit = self._remove_pair_gate(circuit=new_circuit)
        new_circuit = self._combine_rotate_gate(circuit=new_circuit)
        new_circuit = dag_to_circuit(new_circuit, circuit.num)
        return new_circuit

    def _remove_pair_gate(self, circuit: DAGCircuit) -> DAGCircuit:
        """
        This function is to remove two adjacent gates since there are redunant.
        target gate: CX_Gate, H_Gate, CZ_Gate
        """
        sorted_node = self._sorted_node_list(circuit)
        gate_record = {}
        for item in sorted_node.items():
            node = item[1]
            previous_node = self._get_previous_node(node.pos, gate_record)
            if previous_node and previous_node.name in ['CX', 'H', 'CZ'] and node.name == previous_node.name:
                if (previous_node.name != 'CZ' and node.pos == previous_node.pos) or \
                        (previous_node.name == 'CZ' and set(node.pos) == set(previous_node.pos)):
                    # For CX and H doors, the positions must be exactly the same.
                    # For CZ door, the position set must be the same.
                    circuit.remove_instruction_node(previous_node)
                    circuit.remove_instruction_node(node)
                    for pos in node.pos:
                        gate_record[pos].pop()
                    continue

            # Update gate_record
            for pos in node.pos:
                gate_record.setdefault(pos, []).append(node)

        return circuit

    def _combine_rotate_gate(self, circuit: DAGCircuit) -> DAGCircuit:
        """
        This function is to combine two adjacent gate since they can be implemented in one gate
        target gate: RX_Gate, RY_Gate, RZ_Gate
        """
        sorted_node = self._sorted_node_list(circuit)
        previous_node = None
        target_gates = ['RX', 'RZ', 'RY']
        gate_record = {}
        for item in sorted_node.items():
            node = item[1]
            previous_node = self._get_previous_node(node.pos, gate_record)
            if previous_node and previous_node.name in target_gates and (
                    node.name == previous_node.name and node.pos == previous_node.pos):
                new_node = copy.deepcopy(previous_node)

                # new_node.paras = [sum(previous_node.paras + node.paras)] #  sum([int] + [parameter]) is not supported
                # fix:sum([int] + [parameter]) is not supported
                if len(previous_node.paras) == len(node.paras):
                    new_node.paras = [sum([i, j]) for i, j in zip(previous_node.paras, node.paras)]
                else:
                    raise ValueError(
                        'The number of parameters of the two gates is not the same.' + str(previous_node.paras) + str(
                            node.paras))
                # check the type of the new_node.paras
                # print('new_node.paras:', new_node.paras,list(map(type,new_node.paras)))

                predecessors = circuit.node_qubits_predecessors(previous_node)
                successors = circuit.node_qubits_successors(node)
                circuit.remove_instruction_node(previous_node)
                circuit.remove_instruction_node(node)
                circuit.add_instruction_node(new_node, predecessors, successors)
                for pos in node.pos:
                    gate_record[pos].pop()
                    gate_record[pos].append(new_node)
                continue

            # Update gate_record
            for pos in node.pos:
                gate_record.setdefault(pos, []).append(node)

        return circuit

    def _sorted_node_list(self, circuit: DAGCircuit) -> dict:
        node_list = circuit.nodes_dict()
        # print(node_list) # print the node_list
        # for i in list(node_list.keys()):
        #     print(i)
        if 'm' in node_list:
            # fix the first element to be 'm'  or other strange situation

            sorted_node = {k: node_list[k] for k in list(node_list.keys()) if k != 'm'}
            sorted_node = {k: sorted_node[k] for k in sorted(sorted_node)}
            sorted_node['m'] = node_list['m']
        else:
            sorted_node = {k: node_list[k] for k in sorted(node_list)}
        return sorted_node

    def _get_previous_node(self, positions, gate_record):
        """
        Get the previous node based on position, if there are multiple positions,
        make sure the previous node is the same at all positions.
        """
        previous_nodes = {gate_record[pos][-1] for pos in positions if pos in gate_record and gate_record[pos]}
        return previous_nodes.pop() if len(previous_nodes) == 1 else None



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from qsteed.passes.mapping.baselayout import Layout


class DataDict(dict):
    """A default dictionary-like object"""

    def __init__(self, *args, **kwargs):
        super(DataDict, self).__init__(*args, **kwargs)

        self['coupling_list']: List = None
        self['initial_layout']: Layout = None
        self['final_layout']: Layout = None
        self['variables']: List = None

    def __missing__(self, key):
        return None



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from quafu.elements.parameters import Parameter, ParameterExpression

from qsteed.dag.circuit_dag_convert import dag_to_circuit
from qsteed.dag.dagcircuit import DAGCircuit
from qsteed.passes.basepass import BasePass
from functools import lru_cache


class ParaSubstitution(BasePass):
    """
    Substitute the parameters in the circuit with the original parameters
    """

    def __init__(self):
        super().__init__()

        self.para_list = []
        self.para_dict = dict()

    def set_model(self, model):
        """
        Set the model, including information such as backend and parameters.
        here we only need the parameters in the model
        """
        self.para_list = model.get_datadict()['variables']

    def run(self, circuit):
        """
        create a dict to store the original variable parameters and their names as keys, and the original parameters as values
        update  Parameter variables in the circuit gobally, and
        update Parameter variables in the circuit gates,
        return the updated circuit

        Args:
            circuit: the input circuit

        Returns:
            QuantumCircuit: the updated circuit
        """
        if isinstance(circuit, DAGCircuit):
            circuit = dag_to_circuit(circuit, circuit.circuit_qubits)

        para_dict = dict()
        for para in self.para_list:
            para_dict[f'{para.name}'] = para

        self.para_dict = para_dict

        # update the global variables in the circuit
        self.update_global_variables(circuit)

        # update the parameters in the circuit gates
        self.circuit_update_gates_para(circuit)

        circuit.get_parameter_grads()  # this is optional,you can use the function after  all the compilation passes

        # return the updated quantum circuit.if the input is DAGCircuit, return the new quantum circuit
        return circuit

    def update_global_variables(self, circuit):
        """
        update the global variables in the circuit
        """
        circuit._variables = list(self.para_dict.values())

    def circuit_update_gates_para(self, circuit):
        """
        Update all gate parameters in the circuit.
        """

        for gate in circuit.gates:
            gate = self.update_gate_para(gate)

    def update_gate_para(self, gate):
        """
        Update the parameters in the gate.
        """
        for i, para in enumerate(gate.paras):
            if isinstance(para, (Parameter, ParameterExpression)):
                gate.paras[i] = self.update_para_variables(para)
        return gate

    @lru_cache(maxsize=128)
    def update_para_variables(self, para):
        """
        Update the parameters in the Parameter or ParameterExpression variable.
        """
        if isinstance(para, Parameter):
            para = self.para_dict[f'{para.name}']
            return para

        elif isinstance(para, ParameterExpression):
            para.pivot = self.update_para_variables(para.pivot)
            for i, operand in enumerate(para.operands):
                if isinstance(operand, (Parameter, ParameterExpression)):
                    para.operands[i] = self.update_para_variables(operand)

            return para

        else:
            return para


class ParaSubstitutionCached(BasePass):
    """
    Substitute the parameters in the circuit with the original parameters
    """

    def __init__(self):
        super().__init__()
        self.para_list = []
        self.para_dict = dict()
        self._para_dict_cache = None
        self._gate_para_cache = {}

    def set_model(self, model):
        """
        Set the model, including information such as backend and parameters.
        Here we only need the parameters in the model
        """
        self.para_list = model.get_datadict()['variables']
        self._para_dict_cache = None  # Invalidate the cache

    def run(self, circuit):
        """
        Create a dict to store the original variable parameters and their names as keys, and the original parameters as values
        Update Parameter variables in the circuit globally, and
        Update Parameter variables in the circuit gates,
        Return the updated circuit

        Args:
            circuit: the input circuit

        Returns:
            QuantumCircuit: the updated circuit
        """
        if isinstance(circuit, DAGCircuit):
            circuit = dag_to_circuit(circuit, circuit.circuit_qubits)

        if self._para_dict_cache is None:
            para_dict = {f'{para.name}': para for para in self.para_list}
            self.para_dict = para_dict
            self._para_dict_cache = para_dict
        else:
            self.para_dict = self._para_dict_cache

        # Update the global variables in the circuit
        self.update_global_variables(circuit)

        # Update the parameters in the circuit gates
        self.circuit_update_gates_para(circuit)

        circuit.get_parameter_grads()  # This is optional; you can use the function after all the compilation passes

        # Return the updated quantum circuit. If the input is DAGCircuit, return the new quantum circuit
        return circuit

    def update_global_variables(self, circuit):
        """
        Update the global variables in the circuit
        """
        circuit._variables = list(self.para_dict.values())

    def circuit_update_gates_para(self, circuit):
        """
        Update all gate parameters in the circuit.
        """
        for gate in circuit.gates:
            gate = self.update_gate_para(gate)

    def update_gate_para(self, gate):
        """
        Update the parameters in the gate.
        """
        gate_id = id(gate)
        if gate_id not in self._gate_para_cache:
            for i, para in enumerate(gate.paras):
                if isinstance(para, (Parameter, ParameterExpression)):
                    gate.paras[i] = self.update_para_variables(para)
            self._gate_para_cache[gate_id] = gate
        return self._gate_para_cache[gate_id]

    def update_para_variables(self, para):
        """
        Update the parameters in the Parameter or ParameterExpression variable.
        """
        if isinstance(para, Parameter):
            para = self.para_dict[f'{para.name}']
            return para

        elif isinstance(para, ParameterExpression):
            para.pivot = self.update_para_variables(para.pivot)
            for i, operand in enumerate(para.operands):
                if isinstance(operand, (Parameter, ParameterExpression)):
                    para.operands[i] = self.update_para_variables(operand)

            return para

        else:
            return para



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Union

from quafu import QuantumCircuit
from quafu.elements import Instruction

from qsteed.dag.dagcircuit import DAGCircuit


class BasePass(ABC):
    """
    The Metaclass for qsteed transpiler passes.
    """

    @abstractmethod
    def run(self, circuit: Union[QuantumCircuit, DAGCircuit, Instruction]):
        pass


class UnrollPass(BasePass):
    """
    The UnrollPass for qsteed transpiler unroll.
    """

    def __init__(self) -> None:
        self.rule = []
        self.parameter_type = 'constant_gate'  # or 'parameterized_gate'
        self.original = 'cx'
        self.basis = ['cx', 'rx', 'ry', 'rz', 'id']
        self.global_phase = 0

    @abstractmethod
    def run(self, circuit: Union[QuantumCircuit, DAGCircuit, Instruction]):
        pass



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import numpy as np
import scipy.linalg

from qsteed.passes.decomposition.utils import matrix_utils as mu


def demultiplexing(U1, U2):
    """
    U = [U1  0] = [V   0][D  0][W  0]
        [0  U2] = [0   V][0 D*][0  W]
    """
    assert mu.is_unitary(U1)
    assert mu.is_unitary(U2)

    U2d = np.conjugate(U2).T
    x = U1 @ U2d

    # print(x)
    # d, V = np.linalg.eig(U1@U2d)
    D2, V = scipy.linalg.schur(x)
    assert mu.is_diagonal(D2)
    assert mu.is_approx(V @ D2 @ np.conjugate(V).T, U1 @ U2d)

    d_tmp = np.sqrt(np.diag(D2))
    D = np.diag(d_tmp)

    assert mu.is_approx(D @ D, D2)
    Vd = np.conjugate(V).T

    W = D @ Vd @ U2

    assert mu.is_approx(U1, V @ D @ W)
    assert mu.is_approx(U2, V @ (np.conjugate(D).T) @ W)

    ZM = np.zeros_like(W)
    VV = mu.stack_matrices(V, ZM, ZM, V)
    DD = mu.stack_matrices(D, ZM, ZM, np.conjugate(D).T)
    WW = mu.stack_matrices(W, ZM, ZM, W)
    UU = mu.stack_matrices(U1, ZM, ZM, U2)

    # print(isApprox(VV@DD@WW, UU))
    assert mu.is_approx(VV @ DD @ WW, UU)
    assert mu.is_unitary(V)
    return V, D, W


def graycode(n):
    gray_str = []
    for i in range(0, 1 << n):
        gray = i ^ (i >> 1)
        gray_str.append("{0:0{1}b}".format(gray, n))

    return gray_str


def get_multi_control_index(n):
    gray_str = graycode(n)
    size = len(gray_str)

    index_list = []
    for i in range(size):
        str1 = gray_str[i]
        str2 = gray_str[(i + 1) % size]

        tmp = [k for k in range(len(str1)) if str1[k] != str2[k]]
        assert len(tmp) == 1
        index_list.append(tmp[0])

    return index_list


def bitParity(i):
    if i < 2 << 16:
        i = (i >> 16) ^ i
        i = (i >> 16) ^ i
        i = (i >> 8) ^ i
        i = (i >> 4) ^ i
        i = (i >> 2) ^ i
        i = (i >> 1) ^ i
        return i % 2
    else:
        print("Bit parity number too big!")


def genMk2(nqubits):
    genMk_lookuptable = []
    for n in range(1, nqubits + 1):
        size = 2 ** n
        Mk = np.zeros((size, size))
        for i in range(size):
            for j in range(size):
                Mk[i, j] = (-1) ** bitParity(i & j ^ (j >> 1))

        genMk_lookuptable.append(Mk)

    return genMk_lookuptable


def bin2gray(num):
    return num ^ int((num >> 1))


def genMk(k):
    Mk = np.zeros((2 ** k, 2 ** k))
    for i in range(2 ** k):
        for j in range(2 ** k):
            p = i & bin2gray(j)
            strbin = "{0:b}".format(p)
            tmp = [m.start() for m in re.finditer('1', strbin)]
            Mk[i, j] = (-1) ** len(tmp)

    return Mk


def genMk_table(nqubits):
    genMk_lookuptable = []
    for n in range(1, nqubits + 1):
        tmp = genMk(n)
        genMk_lookuptable.append(tmp)

    return genMk_lookuptable



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import CRXGate, CXGate, SGate, RYGate, SdgGate

from qsteed.passes.basepass import UnrollPass


class CRXToCNOT(UnrollPass):
    """The CRXToBasis pass.

    Convert CRX gate to {H, CX, RX}.

    CRX gate decomposition rule:
    q[0] ----*----         q[0] -------*----------------*---------------- 
             |      ≡                  |                |
    q[1] ----RX(θ)----     q[1] --S----+----RY(-θ/2)----+--RY(θ/2)----Sdg--

                               
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = CRXGate(0, 1, 0).name.lower()
        self.basis = [CXGate.name.lower(), SGate.name.lower(), SdgGate.name.lower(), RYGate(0, 0).name.lower()]

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, CRXGate):
            theta = op.paras[0]  # Assuming CRZGate takes one parameter θ

            rule.append(SGate(op.pos[1]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(RYGate(op.pos[1], -theta / 2))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(RYGate(op.pos[1], theta / 2))
            rule.append(SdgGate(op.pos[1]))
        else:
            rule.append(op)
        self.rule = rule
        return rule



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from collections import defaultdict
from math import pi

from quafu import QuantumCircuit
from quafu.elements.element_gates import QuantumGate
from quafu.elements.element_gates import pauli


class RandomCircuit:

    def __init__(self,
                 num_qubit: int = 4,
                 depth: int = None,
                 gates_list: list = None,
                 gates_number: int = None,
                 max_qubit: int = 3,
                 measure: bool = True):
        self.num_qubit = num_qubit
        self.depth = depth
        self.gates_list = gates_list
        self.gates_number = gates_number
        self.max_qubit = max_qubit
        self.measure = measure
        self._circuit = None

    def add_gate(self, gates_group, qubits, used_qubits):
        gate = random.choice(self.gates_list)
        if gate.lower() in gates_group['1-qubit-constant']:
            pos = random.choice(qubits)
            self._circuit.add_gate(QuantumGate.gate_classes[gate](pos))
            used_qubits.add(pos)
        elif gate.lower() in gates_group['1-qubit-parameterized']:
            if gate.lower() != 'u3':
                pos = random.choice(qubits)
                para = random.uniform(0, 2 * pi)
                self._circuit.add_gate(QuantumGate.gate_classes[gate](pos, para))
                used_qubits.add(pos)
            else:
                pos = random.choice(qubits)
                para = [random.uniform(0, 2 * pi) for _ in range(3)]
                self._circuit.add_gate(QuantumGate.gate_classes[gate](pos, para[0], para[1], para[2]))
                used_qubits.add(pos)
        elif gate.lower() in gates_group['2-qubit-constant']:
            pos = random.sample(qubits, 2)
            self._circuit.add_gate(QuantumGate.gate_classes[gate](pos[0], pos[1]))
            [used_qubits.add(i) for i in pos]
        elif gate.lower() in gates_group['2-qubit-parameterized']:
            pos = random.sample(qubits, 2)
            para = random.uniform(0, 2 * pi)
            self._circuit.add_gate(QuantumGate.gate_classes[gate](pos[0], pos[1], para))
            [used_qubits.add(i) for i in pos]
        elif gate.lower() in gates_group['3-qubit-constant']:
            pos = random.sample(qubits, 3)
            self._circuit.add_gate(QuantumGate.gate_classes[gate](pos[0], pos[1], pos[2]))
            [used_qubits.add(i) for i in pos]
        elif gate.lower() in gates_group['multi-qubit-constant']:
            if gate.lower() in ['mcx', 'mcy', 'mcz']:
                pos = random.sample(qubits, self.max_qubit)
                self._circuit.add_gate(QuantumGate.gate_classes[gate](pos[1:], pos[0]))
                [used_qubits.add(i) for i in pos]
            elif gate.lower() in ['mcrx', 'mcry', 'mcrz']:
                pos = random.sample(qubits, self.max_qubit)
                para = random.uniform(0, 2 * pi)
                self._circuit.add_gate(QuantumGate.gate_classes[gate](pos[1:], pos[0], para))
                [used_qubits.add(i) for i in pos]
        else:
            raise NameError("Error: %s gate does not support." % (gate.lower()))

    def random_circuit_gates(self) -> QuantumCircuit:
        if self.gates_list is None:
            self.gates_list = list(QuantumGate.gate_classes.keys())
        gates_group = self.gates_group

        qubits = [q for q in range(self.num_qubit)]
        self._circuit = QuantumCircuit(self.num_qubit)
        used_qubits = set()

        for _ in range(self.gates_number):
            self.add_gate(gates_group, qubits, used_qubits)

        unused_qubits = list(set(qubits) - set(used_qubits))
        if unused_qubits:  # If the qubit is not used, add the IdGate gate.
            for pos in unused_qubits:
                self._circuit.add_gate(pauli.IdGate(pos))
        return self._circuit

    def random_circuit_depth(self) -> QuantumCircuit:
        if self.gates_list is None:
            self.gates_list = list(QuantumGate.gate_classes.keys())

        gates_group = self.gates_group

        qubits = [q for q in range(self.num_qubit)]
        self._circuit = QuantumCircuit(self.num_qubit)
        used_qubits = set()

        depth = 0
        while depth < self.depth:
            self.add_gate(gates_group, qubits, used_qubits)
            depth = len(self._circuit.layered_circuit()[0]) - 1

        unused_qubits = list(set(qubits) - set(used_qubits))
        if unused_qubits:  # If the qubit is not used, add the IdGate gate.
            for pos in unused_qubits:
                self._circuit.add_gate(pauli.IdGate(pos))
        return self._circuit

    def random_circuit(self):
        if self.depth is not None and self.gates_number is None:
            circuit = self.random_circuit_depth()
        elif self.depth is None and self.gates_number is not None:
            circuit = self.random_circuit_gates()
        elif self.depth is not None and self.gates_number is not None:
            circuit = self.random_circuit_gates()
        else:
            raise ValueError("At least one of depth and gates_number must be specified.")
        if self.measure:
            circuit.measure([q for q in range(self.num_qubit)], [q for q in range(self.num_qubit)])
        return circuit

    @property
    def gates_group(self):
        _gates_group = defaultdict(list)
        for gate_name in list(QuantumGate.gate_classes.keys()):
            if gate_name in ['id', 'x', 'y', 'z', 's', 'sdg', 't', 'tdg', 'sx', 'sxdg', 'sy', 'sydg', 'h',
                             'w', 'sw', 'swdg']:
                _gates_group['1-qubit-constant'].append(gate_name)
            elif gate_name in ['rx', 'ry', 'rz', 'u3', 'p']:
                _gates_group['1-qubit-parameterized'].append(gate_name)
            elif gate_name in ['rxx', 'ryy', 'rzz', 'cp', 'crx', 'cry', 'crz']:
                _gates_group['2-qubit-parameterized'].append(gate_name)
            elif gate_name in ['cx', 'cy', 'cz', 'cs', 'ct', 'swap', 'iswap', 'cnot']:
                _gates_group['2-qubit-constant'].append(gate_name)
            elif gate_name in ['ccx', 'cswap', 'toffoli', 'fredon']:
                _gates_group['3-qubit-constant'].append(gate_name)
            elif gate_name in ['mcx', 'mcy', 'mcz', 'mcrx', 'mcry', 'mcrz']:
                _gates_group['multi-qubit-constant'].append(gate_name)
        return _gates_group

# tests
# rqc = RandomCircuit(num_qubit=4, depth=6, gates_number=20, max_qubit=2, gates_list=['cswap'])
# qc = rqc.random_circuit()
# qc.draw_circuit()



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from quafu import QuantumCircuit
from quafu.elements.parameters import Parameter

from qsteed.passes.ParameterTuning.parametersubstitution import ParaSubstitution
from qsteed.passes.mapping.layout.sabre_layout import SabreLayout
from qsteed.passes.optimization.one_qubit_optimization import OneQubitGateOptimization
from qsteed.passes.optimization.optimization_combine import GateCombineOptimization
from qsteed.passes.unroll.unroll_to_2qubit import UnrollTo2Qubit
from qsteed.passes.unroll.unroll_to_basis import UnrollToBasis
from qsteed.passflow.passflow import PassFlow
from qsteed.transpiler.transpiler import Transpiler
from tests.shared_utils import get_initial_model


def order_variables_indices(circuit: QuantumCircuit, initial_variables: list = None):
    circuit.get_parameter_grads()
    transpiled_variables = circuit._variables
    initial_order = {item.name: i for i, item in enumerate(initial_variables)}
    order_indices = [initial_order[x.name] for x in transpiled_variables]
    return order_indices


class TestParameterizedCircuitTranspile:

    def test_parameterized_circuit_transpile(self):
        # Create parameterized circuit
        pq = QuantumCircuit(5)
        theta = [Parameter("theta_%d" % i, 0.1) for i in range(10)]
        theta[4].value = 3.3
        pq.rx(1, 0.8)
        pq.rx(1, theta[2] * theta[4] * theta[3])
        pq.rx(1, 2 * theta[4])
        pq.ry(2, theta[5] + 0.9)
        pq.rx(3, theta[6])
        pq.rx(4, theta[7])
        pq.rxx(0, 3, theta[0])
        pq.rzz(2, 3, theta[8])
        pq.rzz(0, 2, theta[4])
        pq.rx(0, theta[9])
        pq.rx(0, 3.5)
        pq.cx(3, 4)
        pq.ry(2, theta[1].sin() - 4. * theta[0] + theta[2] * theta[0] + 2.5)
        pq.rx(2, theta[1] - 4. * theta[0] + theta[2] * theta[0])
        pq.measure([0, 1, 2, 3, 4], [0, 1, 2, 3, 4])
        print("original circuit:")
        pq.draw_circuit()

        passes = [
            UnrollTo2Qubit(),
            SabreLayout(heuristic='fidelity', max_iterations=3),
            UnrollToBasis(basis_gates=['cx', 'rx', 'ry', 'rz', 'id', 'h']),
            GateCombineOptimization(),
            OneQubitGateOptimization(),
            ParaSubstitution(),  # Parameterized pass
        ]
        passflow = PassFlow(passes=passes)
        initial_model = get_initial_model()

        transpiler = Transpiler(passflow, initial_model)
        transpiled_circuit = transpiler.transpile(pq)
        assert transpiled_circuit is not None
        transpiled_circuit.draw_circuit()

        initial_variables = transpiler.model.datadict['variables']
        assert initial_variables is not None
        print("Initial circuit variables:", initial_variables)
        print("Compiled circuit variables:", transpiled_circuit._variables)

        # After transpilation, the parameters order may be changed,
        # if we want easily update the parameter in the order of the original parameters,
        # we need to get the order map relation between the original parameters and the transpiled parameters
        transpiled_circuit.get_parameter_grads()
        order_indices = order_variables_indices(transpiled_circuit, initial_variables)
        assert order_indices is not None

        update_initial_variables = [0, 0, 0, 9, 0, 0, 0, 0, 0, 0]
        transpiled_circuit._update_params(update_initial_variables, order=order_indices)
        print('Compiled circuit updated parameters:')
        transpiled_circuit.draw_circuit()



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cmath
import math

import numpy as np

from qsteed.passes.decomposition.utils.matrix_utils import get_global_phase


class OneQubitDecompose:
    def __init__(self, method: str = 'ZYZ'):
        self.method = method

    def zyz_decomposition(self, unitary):
        """ZYZ decomposition of arbitrary single-qubit gate (unitary).
        unitary = e^{i*global_phase} @ Rz(gamma) @ Ry(beta) @ Rz(alpha)

        Args:
            unitary (np.array): arbitrary unitary
        Returns:
            global_phase: the global phase of arbitrary unitary
            gamma, beta, alpha: rotation angle
        """
        if unitary.shape[0] == 2:
            global_phase, special_unitary = get_global_phase(unitary)
            beta = 2 * math.atan2(abs(special_unitary[1, 0]), abs(special_unitary[0, 0]))
            t1 = cmath.phase(special_unitary[1, 1])
            t2 = cmath.phase(special_unitary[1, 0])
            alpha = t1 + t2
            gamma = t1 - t2
        else:
            raise Exception("ZYZ decomposition only applies to single-qubit gate.")
        return gamma, beta, alpha, global_phase

    def zxz_decomposition(self, unitary):
        """ZXZ decomposition of arbitrary single-qubit gate (unitary).
        unitary = e^{i*global_phase} @ Rz(gamma) @ Rx(beta) @ Rz(alpha)

        Rx(ϴ) = Rz(-π/2) @ Ry(ϴ) @ Rz(π/2)
         ====>
        Rz(gamma) @ Rx(beta) @ Rz(alpha) = Rz(gamma-π/2) @ Ry(beta) @ Rz(alpha+π/2)

        Args:
            unitary (np.array): arbitrary unitary
        Returns:
            global_phase: the global phase of arbitrary unitary
            gamma, beta, alpha: rotation angle
        """
        if unitary.shape[0] == 2:
            gamma, beta, alpha, global_phase = self.zyz_decomposition(unitary)
        else:
            raise Exception("ZXZ decomposition only applies to single-qubit gate.")
        return gamma - np.pi / 2, beta, alpha + np.pi / 2, global_phase

    def xyx_decomposition(self, unitary):
        """XYX decomposition of arbitrary single-qubit gate (unitary).
        unitary = e^{i*global_phase} @ Rx(gamma) @ Ry(beta) @ Rx(alpha)

        H @ Rx(ϴ) @ H = Rz(ϴ)
        H @ Rz(ϴ) @ H = Rx(ϴ)
        H @ Ry(ϴ) @ H = Ry(-ϴ)
        H @ H = I
         ====>
        Rx(gamma) @ Ry(beta) @ Rx(alpha) = H @ Rz(gamma) @ Ry(-beta) @ Rz(alpha) @ H

        Args:
            unitary (np.array): arbitrary unitary
        Returns:
            global_phase: the global phase of arbitrary unitary
            gamma, beta, alpha: rotation angle
        """
        if unitary.shape[0] == 2:
            hadamard = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
            zyz_form = hadamard @ unitary @ hadamard
            gamma, beta, alpha, global_phase = self.zyz_decomposition(zyz_form)
        else:
            raise Exception("XYX decomposition only applies to single-qubit gate.")
        return gamma, -beta, alpha, global_phase

    def xzx_decomposition(self, unitary):
        """XZX decomposition of arbitrary single-qubit gate (unitary).
        unitary = e^{i*global_phase} @ Rx(gamma) @ Rz(beta) @ Rx(alpha)

        H @ Rx(ϴ) @ H = Rz(ϴ)
        H @ Rz(ϴ) @ H = Rx(ϴ)
        H @ H = I
         ====>
        H @ Rx(gamma) @ Rz(beta) @ Rx(alpha) @ H = Rz(gamma) @ Rx(beta) @ Rz(alpha)

        Args:
            unitary (np.array): arbitrary unitary
        Returns:
            global_phase: the global phase of arbitrary unitary
            gamma, beta, alpha: rotation angle
        """
        if unitary.shape[0] == 2:
            hadamard = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
            zxz_form = hadamard @ unitary @ hadamard
            gamma, beta, alpha, global_phase = self.zxz_decomposition(zxz_form)
        else:
            raise Exception("XZX decomposition only applies to single-qubit gate.")
        return gamma, beta, alpha, global_phase

    def run(self, unitary):
        if self.method == 'ZYZ':
            return self.zyz_decomposition(unitary)
        elif self.method == 'ZXZ':
            return self.zxz_decomposition(unitary)
        elif self.method == 'XYX':
            return self.xyx_decomposition(unitary)
        elif self.method == 'XZX':
            return self.xzx_decomposition(unitary)
        else:
            raise ValueError("The selected decomposition method can only be: ZYZ, ZXZ, XYX, XZX")



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Union, Optional, Dict

from qsteed.graph.couplinggraph import CouplingGraph


class Backend:
    """
    Represents the backend of a quantum system.

    Args:
        name (str): The name of the backend.
        backend_type (str): The type of the backend (e.g., superconducting, ion trap, atomic).
        qubits_num (int): Number of qubits in the backend.
        coupling_map (List[Union[int, tuple]]): The coupling map for the qubits.
        avg_t1 (Optional[float]): The average T1 time.
        avg_t2 (Optional[float]): The average T2 time.
        fidelity (Optional[float]): Fidelity of the backend.
        basis_gates (Optional[List[str]]): List of basis gates supported.
        pulse (Optional[bool]): Indicates if the backend supports pulse level operations.
        calibration_time (Optional[str]): The calibration time.

    Attributes:
        _properties (dict): A dictionary containing all the properties of the backend.
    """

    def __init__(self, name: str = 'backend',
                 backend_type: str = 'superconducting',
                 qubits_num: int = None,
                 qubits_info: List[Dict] = None,
                 coupling_list: List[Union[int, tuple]] = None,
                 coupling_graph: CouplingGraph = None,
                 avg_t1: Optional[float] = None,
                 avg_t2: Optional[float] = None,
                 avg_fidelity: Optional[float] = None,
                 basis_gates: Optional[List[str]] = None,
                 pulse: bool = False,
                 calibration_time: Optional[float] = None,
                 status: str = None,
                 ):

        # Setting up bidirectional coupling graph.
        if coupling_list is not None:
            coupling_graph = CouplingGraph(coupling_list)
            if coupling_graph.is_bidirectional is False:
                coupling_graph.do_bidirectional()
        elif coupling_graph is not None:
            if coupling_graph.is_bidirectional is False:
                coupling_graph.do_bidirectional()
        else:
            print("Warning: Both coupling_list and coupling_graph are None, "
                  "which means there is no qubits coupling structure.")

        self._properties = {
            "name": name,
            "backend_type": backend_type,
            "qubits_num": qubits_num,
            "qubits_info": qubits_info,
            "coupling_list": coupling_list,
            "coupling_graph": coupling_graph,
            "avg_t1": avg_t1,
            "avg_t2": avg_t2,
            "avg_fidelity": avg_fidelity,
            "basis_gates": basis_gates,
            "pulse": pulse,
            "calibration_time": calibration_time,
            "status": status,
        }

    def get_all_properties(self):
        """get all the properties of the backend."""
        return self._properties

    def get_property(self, property_name: str) -> Union[str, int, float, List, None]:
        """Return the value of the specified property."""
        return self._properties.get(property_name, None)

    def set_property(self, property_name: str, value: Union[str, int, float, List]):
        """Set a new value for the specified property."""
        if property_name in self._properties:
            self._properties[property_name] = value

    def __repr__(self):
        return f"<Backend(name={self.get_property('name')}, type={self.get_property('backend_type')}, qubits_num={self.get_property('qubits_num')})> "

    def supports_pulse(self):
        """Check if the backend supports pulse level operations."""
        return self.get_property("pulse")

    def available_basis_gates(self) -> Optional[List[str]]:
        """Return the list of basis gates supported by the backend."""
        return self.get_property("basis_gates")



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cmath

import numpy as np
from quafu.elements.matrices import IdMatrix


def split_matrix(matrix: np.ndarray):
    """
    Evenly split a matrix into 4 sub-matrices.
    """
    top, bottom = np.vsplit(matrix, 2)
    t_left, t_right = np.hsplit(top, 2)
    b_left, b_right = np.hsplit(bottom, 2)
    return t_left, t_right, b_left, b_right


def combine_matrix(topLeft, topRight, bottomLeft, bottomRight):
    top = np.hstack((topLeft, topRight))
    bottom = np.hstack((bottomLeft, bottomRight))
    tot = np.vstack((top, bottom))
    return tot


def stack_matrices(t_left, t_right, b_left, b_right):
    """
    Stack 4 sub-matrices into a matrix.
    """
    top = np.hstack((t_left, t_right))
    bottom = np.hstack((b_left, b_right))
    mat = np.vstack((top, bottom))
    return mat


def is_zero(a):
    return not np.any(np.absolute(a) > 1e-8)


def is_approx(a, b, thres=1e-6):
    # TODO: seems there are some very small elements that cannot be compared correctly
    # if not np.allclose(a, b, rtol=thres, atol=thres):
    #    print(np.sum(a-b))
    return np.allclose(a, b, rtol=thres, atol=thres)


def is_unitary(matrix):
    mat_dg = np.conjugate(matrix).T
    id_mat = np.eye(matrix.shape[0])
    return is_approx(mat_dg @ matrix, id_mat) and is_approx(matrix @ mat_dg, id_mat)


def is_hermitian(matrix):
    tmp = np.conjuate(matrix).T
    return is_approx(tmp, matrix)


def is_diagonal(matrix: np.ndarray):
    diag = np.diag(matrix)
    diag_mat = np.diag(diag)
    return is_approx(matrix, diag_mat)


def is_kron_with_id2(matrix):
    """
    Check if the matrix is a Kronecker product of a matrix and identity matrix.
    """
    nsize = matrix.shape[0]
    a_cond = is_zero(matrix[0:nsize:2, 1:nsize:2])
    b_cond = is_zero(matrix[1:nsize:2, 0:nsize:2])
    c_cond = is_approx(matrix[0, :-1], matrix[1, 1:])
    d_cond = is_approx(matrix[-2, :-1], matrix[-1, 1:])

    return a_cond and b_cond and c_cond and d_cond


def multi_kron(op1, op2, ind1, ind2, nspin):
    tmp = 1
    for i in range(nspin):
        if i == ind1:
            tmp = np.kron(tmp, op1)
        elif i == ind2:
            tmp = np.kron(tmp, op2)
        else:
            tmp = np.kron(tmp, IdMatrix)
    return tmp


def Int2Bas(n, b, nbit):
    if (n == 0): return [0] * nbit
    x = []
    while (n):
        x.append(int(n % b))
        n //= b
    return [0] * (nbit - len(x)) + x[::-1]


def Bas2Int(x, b):
    """
    b: base
    -type: int
    x: the b base bitstring numbers
    -type: list

    example: x = [1, 0, 1], b = 2, return 5
    """
    nbit = len(x)
    z = [b ** (nbit - 1 - i) for i in range(nbit)]
    return np.dot(z, x)


def general_CNOT(nqubit, control_q, target_q):
    CNOT_matrix = np.zeros((2 ** nqubit, 2 ** nqubit))
    for i in range(2 ** nqubit):
        i_bin = Int2Bas(i, 2, nqubit)
        # print("i bin is {}:".format(i_bin))
        if i_bin[control_q] == 0:
            CNOT_matrix[i, i] = 1
        else:
            j_bin = i_bin.copy()
            j_bin[target_q] = (j_bin[target_q] + 1) % 2
            # print("j bin is {}:".format(j_bin))
            j = Bas2Int(j_bin, 2)
            CNOT_matrix[i, j] = 1

    return CNOT_matrix


def general_kron(op, ind, nqubit):
    tmp = 1
    for i in range(nqubit):
        if i == ind:
            tmp = np.kron(tmp, op)
        else:
            tmp = np.kron(tmp, IdMatrix)
    return tmp


def get_global_phase(unitary):
    """ Get the global phase of arbitrary unitary, and get the special unitary.

    Args:
        unitary (np.array): arbitrary unitary
    Returns:
        global_phase: the global phase of arbitrary unitary
        special_unitary (np.array)
    """
    coefficient = np.linalg.det(unitary) ** (-0.5)
    global_phase = -cmath.phase(coefficient)
    special_unitary = coefficient * unitary
    return global_phase, special_unitary


def matrix_distance_squared(u1, u2):
    """Used to compare the distance of two matrices. The global phase is ignored.

    Args:
        u1 (np.array): A unitary matrix in the form of a numpy ndarray.
        u2 (np.array): Another unitary matrix.

    Returns:
        Float : A single value between 0 and 1. A value close to 0 indicates that u1 and u2 are the same unitary.
    """
    return np.abs(1 - np.abs(np.sum(np.multiply(u1, np.conj(u2)))) / u1.shape[0])



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
from quafu import QuantumCircuit
from quafu.elements.matrices import ry_mat, rz_mat, CXMatrix, rx_mat

from qsteed.passes.decomposition.CSD_decompose import fatCSD
from qsteed.passes.decomposition.one_qubit_decompose import OneQubitDecompose
from qsteed.passes.decomposition.utils import decompose_utils as du
from qsteed.passes.decomposition.utils import matrix_utils as mu


class UnitaryDecompose(object):
    def __init__(self, array, qubits, endian: str = None, one_qubit_decompose: str = 'XZX'):
        """
        Args:
            array (np.array): arbitrary unitary
            qubits (list): example [0,1,2,3]
            endian (str): 'little-endian': U_qn ⊗ ... ⊗ U_q1 ⊗ U_q0, where q0<q1< ... <qn
                          'big-endian': U_q0 ⊗ U_q1 ⊗ ... ⊗ U_qn, where q0<q1< ... <qn
            one_qubit_decompose (str): one-qubit decomposition method: 'ZYZ', 'ZXZ', 'XYX', 'XZX'
        """
        if endian is None or endian == 'big-endian':
            # big-endian mode: U_q0 ⊗ U_q1 ⊗ ... ⊗ U_qn, where q0<q1< ... <qn
            self.endian = 'big-endian'
        elif endian == 'little-endian':
            self.endian = 'little-endian'
            # little-endian mode: U_qn ⊗ ... ⊗ U_q1 ⊗ U_q0, where q0<q1< ... <qn
            # If you want to use little-endian mode, use this command to invert qubits
            qubits = qubits[::-1]
        else:
            self.endian = 'Warning: endian can only be big-endian or little-endian'
            print('Warning: endian can only be big-endian or little-endian')
        self.array = array
        self.nqubit = len(qubits)
        self.qubits = qubits
        self.one_qubit_decompose = one_qubit_decompose

        self.Mk_table = du.genMk_table(self.nqubit)  # initialize the general M^k lookup table
        self.gates_list = []
        self.quafuQC = QuantumCircuit(self.nqubit)
        self.global_phase = 0

    def decompose(self):
        _matrix = self.array
        self._decompose_matrix(_matrix, self.qubits)
        self.quafuQC.measure(self.qubits, self.qubits)
        # print("Done decomposing")

    def _decompose_matrix(self, _matrix, qubits):
        # qubits = qubits[::-1]
        # tests the matrix is unitary
        assert mu.is_unitary(_matrix)
        num_qubit = len(qubits)
        assert int(np.log2(_matrix.shape[0])) == num_qubit

        if num_qubit == 1:
            # self.gates_list.append((_matrix, qubits, 'U'))
            one_qubit_decomposer = OneQubitDecompose(method=self.one_qubit_decompose)
            gamma, beta, alpha, global_phase = one_qubit_decomposer.run(_matrix)
            self.global_phase += global_phase
            self.one_qubit_circuit(gamma, beta, alpha, qubits)
            # if self.one_qubit_decompose == 'ZYZ':
            #     # # ZYZ decomposition for single-qubit gate
            #     # gamma, beta, alpha, global_phase = zyz_decomposition(_matrix)
            #     # gamma, beta, alpha, global_phase = one_qubit_decomposer.run(_matrix)
            #     # self.global_phase += global_phase
            #     self.gates_list.append((rz_mat(gamma), qubits, 'RZ', gamma))
            #     self.gates_list.append((ry_mat(beta), qubits, 'RY', beta))
            #     self.gates_list.append((rz_mat(alpha), qubits, 'RZ', alpha))
            #     self.quafuQC.rz(qubits[0], gamma)
            #     self.quafuQC.ry(qubits[0], beta)
            #     self.quafuQC.rz(qubits[0], alpha)

            # # # ZXZ decomposition for single-qubit gate
            # gamma, beta, alpha, global_phase = zxz_decomposition(_matrix)
            # self.global_phase += global_phase
            # self.gates_list.append((rz_mat(gamma), qubits, 'RZ', gamma))
            # self.gates_list.append((rx_mat(beta), qubits, 'RX', beta))
            # self.gates_list.append((rz_mat(alpha), qubits, 'RZ', alpha))
            # self.quafuQC.rz(qubits[0], gamma)
            # self.quafuQC.rx(qubits[0], beta)
            # self.quafuQC.rz(qubits[0], alpha)

            # # XYX decomposition for single-qubit gate
            # gamma, beta, alpha, global_phase = xyx_decomposition(_matrix)
            # self.global_phase += global_phase
            # self.gates_list.append((rx_mat(gamma), qubits, 'RX', gamma))
            # self.gates_list.append((ry_mat(beta), qubits, 'RY', beta))
            # self.gates_list.append((rx_mat(alpha), qubits, 'RX', alpha))
            # self.quafuQC.rx(qubits[0], gamma)
            # self.quafuQC.ry(qubits[0], beta)
            # self.quafuQC.rx(qubits[0], alpha)

            # # XZX decomposition for single-qubit gate
            # gamma, beta, alpha, global_phase = xzx_decomposition(_matrix)
            # self.global_phase += global_phase
            # self.gates_list.append((rx_mat(gamma), qubits, 'RX', gamma))
            # self.gates_list.append((rz_mat(beta), qubits, 'RZ', beta))
            # self.gates_list.append((rx_mat(alpha), qubits, 'RX', alpha))
            # self.quafuQC.rx(qubits[0], gamma)
            # self.quafuQC.rz(qubits[0], beta)
            # self.quafuQC.rx(qubits[0], alpha)
        else:
            U00, U01, U10, U11 = mu.split_matrix(_matrix)

            # if bottomLeftCorner(n)==0 and topRightCorner(n)==0:
            if mu.is_zero(U01) and mu.is_zero(U10):
                if mu.is_approx(U00, U11):
                    # print('Optimization: Unitaries are equal, skip one step in the recursion for unitaries of size')
                    self._decompose_matrix(U00, qubits[1:])
                else:
                    # print("Optimization: q2 is zero, only demultiplexing will be performed.")
                    V, D, W = du.demultiplexing(U00, U11)

                    self._decompose_matrix(W, qubits[1:])
                    self.multi_controlled_z(D, qubits[1:], qubits[0])
                    self._decompose_matrix(V, qubits[1:])
            # Check to see if it the kronecker product of a bigger matrix and the identity matrix.
            # By checking if the first row is equal to the second row one over, and if the last two rows are equal
            # Which means the last qubit is not affected by this gate
            elif mu.is_kron_with_id2(_matrix):
                # print("The last qubits not affect.")
                nsize = len(_matrix)
                self._decompose_matrix(_matrix[0:nsize:2, 0:nsize:2], qubits[:-1])

            else:
                # print("CSD decomposition.")
                L0, L1, R0, R1, c, s = fatCSD(_matrix)
                V, D, W = du.demultiplexing(R0, R1)

                self._decompose_matrix(W, qubits[1:])
                self.multi_controlled_z(D, qubits[1:], qubits[0])
                self._decompose_matrix(V, qubits[1:])

                self.multi_controlled_y(s, qubits[1:], qubits[0])

                V, D, W = du.demultiplexing(L0, L1)
                self._decompose_matrix(W, qubits[1:])
                self.multi_controlled_z(D, qubits[1:], qubits[0])
                self._decompose_matrix(V, qubits[1:])

    def multi_controlled_z(self, D, qubits, target_qubit):
        assert len(qubits) == int(math.log(D.shape[0], 2))
        num_qubit = len(qubits)
        # print("The size of D matrix {}".format(len(qubits)))
        # print(qubits)

        # alphas = -2*1j*np.log(np.diag(D))
        alphas = 2 * 1j * np.log(np.diag(D))
        Mk = self.Mk_table[num_qubit - 1]
        thetas = np.linalg.solve(Mk, alphas)
        # print('thetas',thetas)
        thetas = thetas.real
        assert len(thetas) == 2 ** num_qubit

        index = du.get_multi_control_index(num_qubit)
        assert len(index) == len(thetas)

        for i in range(len(index)):
            control_qubit = qubits[index[i]]
            self.gates_list.append((rz_mat(thetas[i]), [target_qubit], 'RZ', thetas[i]))
            self.quafuQC.rz(target_qubit, thetas[i])
            self.gates_list.append((CXMatrix, [control_qubit, target_qubit], 'CX'))
            self.quafuQC.cnot(control_qubit, target_qubit)

    def multi_controlled_y(self, ss, qubits, target_qubit):
        assert len(qubits) == int(math.log(ss.shape[0], 2))
        num_qubit = len(qubits)

        alphas = -2 * np.arcsin(np.diag(ss))
        Mk = self.Mk_table[num_qubit - 1]
        thetas = np.linalg.solve(Mk, alphas)
        # print(thetas)
        thetas = thetas.real
        assert len(thetas) == 2 ** num_qubit

        index = du.get_multi_control_index(num_qubit)
        assert len(index) == len(thetas)

        for i in range(len(index)):
            control_qubit = qubits[index[i]]
            self.gates_list.append((ry_mat(thetas[i]), [target_qubit], 'RY', thetas[i]))
            self.quafuQC.ry(target_qubit, thetas[i])
            self.gates_list.append((CXMatrix, [control_qubit, target_qubit], 'CX'))
            self.quafuQC.cnot(control_qubit, target_qubit)

    def one_qubit_circuit(self, gamma, beta, alpha, qubits):
        if self.one_qubit_decompose == 'ZYZ':
            # ZYZ decomposition for single-qubit gate
            self.gates_list.append((rz_mat(gamma), qubits, 'RZ', gamma))
            self.gates_list.append((ry_mat(beta), qubits, 'RY', beta))
            self.gates_list.append((rz_mat(alpha), qubits, 'RZ', alpha))
            self.quafuQC.rz(qubits[0], gamma)
            self.quafuQC.ry(qubits[0], beta)
            self.quafuQC.rz(qubits[0], alpha)
        elif self.one_qubit_decompose == 'ZXZ':
            # ZXZ decomposition for single-qubit gate
            self.gates_list.append((rz_mat(gamma), qubits, 'RZ', gamma))
            self.gates_list.append((rx_mat(beta), qubits, 'RX', beta))
            self.gates_list.append((rz_mat(alpha), qubits, 'RZ', alpha))
            self.quafuQC.rz(qubits[0], gamma)
            self.quafuQC.rx(qubits[0], beta)
            self.quafuQC.rz(qubits[0], alpha)
        elif self.one_qubit_decompose == 'XYX':
            # XYX decomposition for single-qubit gate
            self.gates_list.append((rx_mat(gamma), qubits, 'RX', gamma))
            self.gates_list.append((ry_mat(beta), qubits, 'RY', beta))
            self.gates_list.append((rx_mat(alpha), qubits, 'RX', alpha))
            self.quafuQC.rx(qubits[0], gamma)
            self.quafuQC.ry(qubits[0], beta)
            self.quafuQC.rx(qubits[0], alpha)
        elif self.one_qubit_decompose == 'XZX':
            # XZX decomposition for single-qubit gate
            self.gates_list.append((rx_mat(gamma), qubits, 'RX', gamma))
            self.gates_list.append((rz_mat(beta), qubits, 'RZ', beta))
            self.gates_list.append((rx_mat(alpha), qubits, 'RX', alpha))
            self.quafuQC.rx(qubits[0], gamma)
            self.quafuQC.rz(qubits[0], beta)
            self.quafuQC.rx(qubits[0], alpha)
        else:
            raise ValueError("The selected decomposition method can only be: ZYZ, ZXZ, XYX, XZX")



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import pi
from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import CTGate, CXGate, PhaseGate

from qsteed.passes.basepass import UnrollPass


class CTToCNOT(UnrollPass):
    """The CTToCNOT pass.

    convert CT gate to {CNOT, Phase}.

    CT gate decomposition rule:
    q[0]  -----*------     q[0]  --P(π/8)----*---------------*-------------
               |        ≡                    |               |
    q[1]  ---P(π/4)---     q[1]  ------------+----P(-π/8)----+----P(π/8)---
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = CTGate.name.lower()
        self.basis = [CXGate.name.lower(), PhaseGate(0, 0).name.lower()]

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, CTGate):
            rule.append(PhaseGate(op.pos[0], pi / 8))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(PhaseGate(op.pos[1], -pi / 8))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(PhaseGate(op.pos[1], pi / 8))
        else:
            rule.append(op)
        self.rule = rule
        return rule



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cmath
import math

import numpy as np

from qsteed.passes.decomposition.utils.matrix_utils import get_global_phase


def zyz_decomposition(unitary):
    """ZYZ decomposition of arbitrary single-qubit gate (unitary).
    unitary = e^{i*global_phase} @ Rz(gamma) @ Ry(beta) @ Rz(alpha)

    Args:
        unitary (np.array): arbitrary unitary
    Returns:
        global_phase: the global phase of arbitrary unitary
        gamma, beta, alpha: rotation angle
    """
    if unitary.shape[0] == 2:
        global_phase, special_unitary = get_global_phase(unitary)
        beta = 2 * math.atan2(abs(special_unitary[1, 0]), abs(special_unitary[0, 0]))
        t1 = cmath.phase(special_unitary[1, 1])
        t2 = cmath.phase(special_unitary[1, 0])
        alpha = t1 + t2
        gamma = t1 - t2
    else:
        raise Exception("ZYZ decomposition only applies to single-qubit gate.")
    return gamma, beta, alpha, global_phase


def zxz_decomposition(unitary):
    """ZXZ decomposition of arbitrary single-qubit gate (unitary).
    unitary = e^{i*global_phase} @ Rz(gamma) @ Rx(beta) @ Rz(alpha)

    Rx(ϴ) = Rz(-π/2) @ Ry(ϴ) @ Rz(π/2)
     ====>
    Rz(gamma) @ Rx(beta) @ Rz(alpha) = Rz(gamma-π/2) @ Ry(beta) @ Rz(alpha+π/2)

    Args:
        unitary (np.array): arbitrary unitary
    Returns:
        global_phase: the global phase of arbitrary unitary
        gamma, beta, alpha: rotation angle
    """
    if unitary.shape[0] == 2:
        gamma, beta, alpha, global_phase = zyz_decomposition(unitary)
    else:
        raise Exception("ZXZ decomposition only applies to single-qubit gate.")
    return gamma - np.pi / 2, beta, alpha + np.pi / 2, global_phase


def xyx_decomposition(unitary):
    """XYX decomposition of arbitrary single-qubit gate (unitary).
    unitary = e^{i*global_phase} @ Rx(gamma) @ Ry(beta) @ Rx(alpha)

    H @ Rx(ϴ) @ H = Rz(ϴ)
    H @ Rz(ϴ) @ H = Rx(ϴ)
    H @ Ry(ϴ) @ H = Ry(-ϴ)
    H @ H = I
     ====>
    Rx(gamma) @ Ry(beta) @ Rx(alpha) = H @ Rz(gamma) @ Ry(-beta) @ Rz(alpha) @ H

    Args:
        unitary (np.array): arbitrary unitary
    Returns:
        global_phase: the global phase of arbitrary unitary
        gamma, beta, alpha: rotation angle
    """
    if unitary.shape[0] == 2:
        hadamard = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
        zyz_form = hadamard @ unitary @ hadamard
        gamma, beta, alpha, global_phase = zyz_decomposition(zyz_form)
    else:
        raise Exception("XYX decomposition only applies to single-qubit gate.")
    return gamma, -beta, alpha, global_phase


def xzx_decomposition(unitary):
    """XZX decomposition of arbitrary single-qubit gate (unitary).
    unitary = e^{i*global_phase} @ Rx(gamma) @ Rz(beta) @ Rx(alpha)

    H @ Rx(ϴ) @ H = Rz(ϴ)
    H @ Rz(ϴ) @ H = Rx(ϴ)
    H @ H = I
     ====>
    H @ Rx(gamma) @ Rz(beta) @ Rx(alpha) @ H = Rz(gamma) @ Rx(beta) @ Rz(alpha)

    Args:
        unitary (np.array): arbitrary unitary
    Returns:
        global_phase: the global phase of arbitrary unitary
        gamma, beta, alpha: rotation angle
    """
    if unitary.shape[0] == 2:
        hadamard = np.array([[1, 1], [1, -1]]) / np.sqrt(2)
        zxz_form = hadamard @ unitary @ hadamard
        gamma, beta, alpha, global_phase = zxz_decomposition(zxz_form)
    else:
        raise Exception("XZX decomposition only applies to single-qubit gate.")
    return gamma, beta, alpha, global_phase



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import matplotlib.pyplot as plt


def plot_probabilities(sampling_dict: dict, complete_bitstring: bool = True, plot_type: str = "probability"):
    """
    Plot the probabilities or counts from execute results.

    Args:
        sampling_dict (dict): Dictionary with bitstring keys and counts/probabilities as values.
        complete_bitstring (bool): If True, include all possible bitstrings in the plot,
                                    even if they have zero counts/probabilities.
        plot_type (str): "probability" to plot probabilities, "count" to plot raw counts.
    """
    if sampling_dict is None:
        raise ValueError("sampling_dict cannot be None")

    # Detect if the sampling_dict contains counts or probabilities
    values = list(sampling_dict.values())
    is_probability = all(0 <= abs(value) <= 1.9 for value in values) and abs(sum(values) - 1) < 9e-1

    qubits = len(list(sampling_dict.keys())[0])
    total_counts = sum(sampling_dict.values())

    if complete_bitstring:
        all_keys = [format(i, f'0{qubits}b') for i in range(2 ** qubits)]
        complete_dict = {key: sampling_dict.get(key, 0) for key in all_keys}
    else:
        complete_dict = sampling_dict

    # Sort the dictionary by key
    sorted_items = sorted(complete_dict.items())
    bitstrs = [item[0] for item in sorted_items]
    values = [item[1] for item in sorted_items]

    probs = values
    ylabel = "Probabilities"
    if plot_type == "probability":
        if not is_probability:
            probs = [value / total_counts for value in values]
    elif plot_type == "count":
        if is_probability:
            ylabel = "Probabilities"
            warnings.warn("Input sampling_dict data is probabilities.", Warning)
        else:
            ylabel = "Counts"

    plt.figure()
    plt.bar(range(len(probs)), probs, tick_label=bitstrs)
    plt.xticks(rotation=70)
    plt.xlabel("Bitstrings")
    plt.ylabel(ylabel)
    plt.tight_layout()
    plt.show()



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

import networkx as nx
from quafu import QuantumCircuit

from qsteed.graph.circuitgraph import circuit_to_graph, relabel_graph
from qsteed.graph.graphkernel import wl_subtree_kernel


def similar_struct(circuit: QuantumCircuit, sub_data):
    """ Compare the similarity of weighted graph of circuit and substructure of quantum chip.

    Args:
        circuit(QuantumCircuit)
        sub_data: backend library data e.g. substructure_data['substructure_dict']
    Returns:
        most_similar_struct(list): Chip substructures similar to quantum circuits,
                                    sorted according to the value of the kernel,
                                    the higher the kernel value, the more similar.
    """
    g1 = circuit_to_graph(circuit)
    similar_struct_list = []
    kernel_value_list = []
    cut = 100  # number of substructures to search
    k = 0
    for sub_list in sub_data:
        if k < cut:
            g2 = nx.Graph()
            g1_copy = copy.deepcopy(g1)
            for item in sub_list:
                g2.add_edges_from([(item[0], item[1], {'weight': item[2]})])
            g2 = relabel_graph(g2)

            # # W-L subtree kernel iteration
            kernel_value = wl_subtree_kernel(g1_copy, g2, iteration=5)
            if round(kernel_value, 0) not in kernel_value_list:
                kernel_value_list.append(round(kernel_value, 0))
                similar_struct_list.append((sub_list, kernel_value))

            # # fast subtree kernel iteration
            # g2 = nx.convert_node_labels_to_integers(g2)
            # kernel_value = fast_subtree_kernel(g1_copy, g2, iteration=2)
            # if round(kernel_value, 0) not in kernel_value_list:
            #     kernel_value_list.append(round(kernel_value, 0))
            #     similar_struct_list.append((sub_list, kernel_value))

            g1_copy.clear()
            g2.clear()
            k += 1

    similar_struct_list = sorted(similar_struct_list, key=lambda x: x[1], reverse=True)
    return similar_struct_list



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List

from quafu import QuantumCircuit

from qsteed.dag.circuit_dag_convert import circuit_to_dag, dag_to_circuit, copy_dag
from qsteed.dag.dagcircuit import DAGCircuit
from qsteed.graph.couplinggraph import CouplingGraph
from qsteed.passes.basepass import BasePass
from qsteed.passes.datadict import DataDict
from qsteed.passes.mapping.baselayout import Layout
from qsteed.passes.mapping.routing.sabre_routing import SabreRouting
from qsteed.utils.reverse_circuit import reverse_circuit


class SabreLayout(BasePass):
    """In order to perform a "SABRE" on any given circuit, it needs to be iterated multiple times
     until convergence or a maximum number of iterations is reached, resulting in the final layout.

     References:
        Gushu Li, Yufei Ding, and Yuan Xie. "Tackling the qubit mapping problem
        for NISQ-era quantum devices." ASPLOS ’19, 1001–1014 (ACM, 2019).
        URL https://doi.org/10.1145/3297858.3304023
    """

    def __init__(self, coupling_list: List = None,
                 heuristic="distance",
                 routing_pass=None,
                 max_iterations=3,
                 sabre_initial_layout=None):
        """SabreLayout initializer.

        Args:
            coupling_list (List): qubit coupling structure, e.g. [(0,1,0.99), (1,0,0.99), (1,2,0.98),(2,1,0.95)]
            heuristic (str): 'distance': The original SABER algorithm is based on the swap gates distance;
                             'fidelity': Noise-aware SABER algorithm based on two-qubit gates fidelity;
                             'mixture': First based on the distance, when the distance cannot be determined,
                                        based on the two-qubit gate fidelity.
            routing_pass (BasePass): the routing pass to use while iterating.
            max_iterations (int): number of forward-backward iterations.
                                  If max_iterations=0, only perform one forward SabreRouting
            sabre_initial_layout (Layout): The initial layout of the SABER algorithm.
                                           A suitable initial layout can reduce the number of SWAP gates.
        """
        super().__init__()

        self.coupling_list = coupling_list
        self.coupling_graph = None
        self.routing_pass = routing_pass
        self.max_iterations = max_iterations
        self.sabre_initial_layout = sabre_initial_layout
        self.model = None
        self.heuristic = heuristic

    def set_model(self, model):
        """Set the model, including information such as backend and layout.

        Args:
            model (Model): The given model includes information such as backend and layout.
        """
        self.model = model
        # self.coupling_list = self.model.get_backend().get_property('coupling_list')
        self.coupling_graph = self.model.get_backend().get_property('coupling_graph')

        if self.model.datadict is None:
            self.model.datadict = DataDict()

    def get_model(self):
        """Get the model, including information such as backend and layout.
        """
        return self.model

    def run(self, circ_dag: Union[QuantumCircuit, DAGCircuit]):
        """Run the SabreLayout pass on DAGCircuit.

        Args:
            circ_dag (DAGCircuit or QuantumCircuit): Find the layout of this circuit.

        Returns:
            new_circuit (QuantumCircuit): Quantum circuit satisfying hardware coupling structure.
        """
        if isinstance(circ_dag, QuantumCircuit):
            dag = circuit_to_dag(circ_dag)
            # circuit = copy.deepcopy(circ_dag)
            circuit = circ_dag
        elif isinstance(circ_dag, DAGCircuit):
            # dag = copy.deepcopy(circ_dag)
            dag = copy_dag(circ_dag)
            circuit = dag_to_circuit(circ_dag, circ_dag.circuit_qubits)
        else:
            raise TypeError('Error: SabreLayout pass only supports QuantumCircuit or DAGCircuit.')

        if len(dag.qubits_used) == 1:  # Single-qubit circuit do not need to perform sabre
            return circuit

        if self.coupling_graph is None:
            if self.coupling_list is not None:
                coupling_graph = CouplingGraph(self.coupling_list)
                if coupling_graph.is_bidirectional is False:
                    coupling_graph.do_bidirectional()
                self.coupling_graph = coupling_graph
            else:
                raise ValueError("Error: There is no qubits coupling structure.")

        if len(dag.qubits_used) > self.coupling_graph.num_qubits:
            raise ValueError("More virtual qubits than physical qubits.")

        measure_nodes = dag.get_measure_nodes()
        dag.remove_measure_nodes()

        if self.model.get_layout()["final_layout"] is None:
            layout = Layout()
            # Method1: Choose a trivial initial_layout.
            # layout.generate_trivial_layout(virtual_qubits=dag.circuit_qubits)
            # Method2: Choose a random initial_layout.
            layout.generate_random_layout(len(dag.qubits_used), self.coupling_graph.num_qubits)
            # TODO: The required qubits are less than the number of physical qubits
            # Complete mapping, even if the actual number of qubits used in the line is less than the hardware qubits.
            # if len(layout.p2v) < self.coupling_graph.num_qubits:
            self.model.set_layout({'initial_layout': layout})

        if self.routing_pass is None:
            self.routing_pass = SabreRouting(heuristic=self.heuristic, modify_dag=False)
        else:
            self.routing_pass.modify_dag = False

        rev_circuit = reverse_circuit(circuit)
        rev_dag = circuit_to_dag(rev_circuit)

        # Do forward-backward iterations.
        for _ in range(self.max_iterations):
            for _ in ("forward", "backward"):
                self.run_single(dag)
                # import copy
                # self.model.set_layout({'initial_layout': copy.deepcopy(self.model.get_layout()['final_layout'])})
                self.model.set_layout({'initial_layout': self.model.get_layout()['final_layout']})
                dag, rev_dag = rev_dag, dag
                # dag, rev_dag = copy.deepcopy(rev_dag), copy.deepcopy(dag)

        # The last forward iteration obtains the final circuit.
        self.routing_pass.modify_dag = True
        self.nodes_label = []
        physical_dag = self.run_single(dag)
        physical_circuit = dag_to_circuit(physical_dag, physical_dag.circuit_qubits)

        # Add measurement, i.e. mapping of physical qubits to classical qubits
        p2c = {}  # mapping for physical qubits to classic qubits
        for virtual, classic in measure_nodes[-1].pos.items():
            p2c[self.model.get_layout()['final_layout'].v2p[virtual]] = classic
        p2c = dict(sorted(p2c.items(), key=lambda item: item[1]))
        physical_circuit.measure(list(p2c.keys()), list(p2c.values()))

        return physical_circuit

    def run_single(self, dag):
        """Execute a single sabre routing.

        Args:
            dag (DAGCircuit): dag or rev_dag for iteration.

        Returns:
            new_dag (DAGCircuit): The new dag obtained by executing the sabre routing algorithm once.
        """
        if hasattr(self.routing_pass, 'set_model'):
            self.routing_pass.set_model(self.model)

        new_dag = self.routing_pass.run(dag)

        if hasattr(self.routing_pass, 'get_model'):
            self.model = self.routing_pass.get_model()
        return new_dag



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import numpy as np
from scipy.linalg import expm

# Helper Functions

I = np.identity(2)
X = np.array([[0, 1], [1, 0]])
Y = np.array([[0, -1j], [1j, 0]])
Z = np.array([[1, 0], [0, -1]])


def decompose_one_qubit_product(
        U: np.ndarray, validate_input: bool = True, atol: float = 1e-8, rtol: float = 1e-5
):
    """
    Decompose a 4x4 unitary matrix to two 2x2 unitary matrices.
    Args:
        U (np.ndarray): input 4x4 unitary matrix to decompose.
        validate_input (bool): if check input.
    Returns:
        phase (float): global phase.
        U1 (np.ndarray): decomposed unitary matrix U1.
        U2 (np.ndarray): decomposed unitary matrix U2.
        atol (float): absolute tolerance of loss.
        rtol (float): relative tolerance of loss.
    Raises:
        AssertionError: if the input is not a 4x4 unitary or
        cannot be decomposed.
    """

    """if validate_input:
        assert np.allclose(
            makhlin_invariants(U, atol=atol, rtol=rtol), (1, 0, 3), atol=atol, rtol=rtol
        )"""

    # find the maximum value's index of U
    i, j = np.unravel_index(np.argmax(U, axis=None), U.shape)

    def u1_set(i):
        return (1, 3) if i % 2 else (0, 2)

    def u2_set(i):
        return (0, 1) if i < 2 else (2, 3)

    # get the submatrix
    u1 = U[np.ix_(u1_set(i), u1_set(j))]
    u2 = U[np.ix_(u2_set(i), u2_set(j))]

    u1 = to_su(u1)
    u2 = to_su(u2)

    phase = U[i, j] / (u1[i // 2, j // 2] * u2[i % 2, j % 2])

    return phase, u1, u2


def to_su(u: np.ndarray) -> np.ndarray:
    """
    Given a unitary in U(N), return the
    unitary in SU(N).
    Args:
        u (np.ndarray): The unitary in U(N).
    Returns:
        np.ndarray: The unitary in SU(N)
    """
    # turn the determinent to 1
    return u * complex(np.linalg.det(u)) ** (-1 / np.shape(u)[0])


def weyl_chamber(c):
    """Bring coordinates vector into the Weyl chamber"""

    # Step 0: work in terms of multiple of pi
    c /= np.pi

    # Step 1: Bring everything into [0, 1)
    c -= np.floor(c)

    # Step 2: Sort c1 >= c2 >= c3
    c = np.sort(c)[::-1]

    # Step 3: if c1 + c2 >= 1, transform (c1, c2, c3) -> (1-c2, 1-c1, c3)
    if c[0] + c[1] >= 1:
        c = np.sort(np.array([1 - c[1], 1 - c[0], c[2]]))[::-1]

    # Step 4: if c3 = 0 and c1>1/2, transform (c1, c2, 0) -> (1-c1, c2, 0)
    if (c[0] > 1 / 2) and np.isclose(c[2], 0):
        c = np.array([1 - c[0], c[1], 0])

    # Step 5: Turn it back into radians
    c *= np.pi

    return c


def KAK_decomposition(
        U: np.ndarray,
        rounding: int = 19
) -> Tuple[float, np.ndarray, np.ndarray, float, np.ndarray, np.ndarray, float,
           float, float]:
    """
    Decomposes a 2-qubit unitary matrix into the product of three matrices:
    KAK = L @ CAN(theta_vec) @ R where L and R are two-qubit local unitaries, 
    CAN is a 3-parameter canonical matrix, and theta_vec is a vector of 3 angles.

    Args:
        U (np.ndarray): 2-qubit unitary matrix
        rounding (int): Number of decimal places to round intermediate 
        matrices to (default 14)

    Returns:
        Tuple of 9 values:
            - phase1 (float): Global phase factor for left local unitary L
            - L1 (np.ndarray): Top 2x2 matrix of left local unitary L
            - L2 (np.ndarray): Bottom 2x2 matrix of left local unitary L
            - phase2 (float): Global phase factor for right local unitary R
            - R1 (np.ndarray): Top 2x2 matrix of right local unitary R
            - R2 (np.ndarray): Bottom 2x2 matrix of right local unitary R
            - c0 (float): XX canonical parameter in the Weyl chamber
            - c1 (float): YY canonical parameter in the Weyl chamber
            - c2 (float): ZZ canonical parameter in the Weyl chamber
    """

    # 0. Map U(4) to SU(4) (and phase)
    global_phase = np.linalg.det(U) ** 0.25
    U /= global_phase

    assert np.isclose(np.linalg.det(U), 1), "Determinant of U is not 1"

    # 1. Unconjugate U into the magic basis
    B = 1 / np.sqrt(2) * np.array([[1, 0, 0, 1j], [0, 1j, 1, 0],
                                   [0, 1j, -1, 0], [1, 0, 0, -1j]])  # Magic Basis
    U_prime = np.conj(B).T @ U @ B

    Theta = lambda U: np.conj(U)
    # Isolating the maximal torus
    M_squared = Theta(np.conj(U_prime).T) @ U_prime

    if rounding is not None:
        M_squared = np.round(M_squared, rounding)  # For numerical stability

    ## 2. Diagonalizing M^2
    D, P = np.linalg.eig(M_squared)

    ## Check and correct for det(P) = -1
    if np.isclose(np.linalg.det(P), -1):
        P[:, 0] *= -1  # Multiply the first eigenvector by -1

    # 3. Extracting K2
    K2 = np.conj(P).T

    assert np.allclose(K2 @ K2.T, np.identity(4)), "K2 is not orthogonal"
    assert np.isclose(np.linalg.det(K2), 1), "Determinant of K2 is not 1"

    # 4. Extracting A
    A = np.sqrt(D)

    ## Check and correct for det(A) = -1
    if np.isclose(np.prod(A), -1):
        A[0] *= -1  # Multiply the first eigenvalue by -1

    A = np.diag(A)  # Turn the list of eigenvalues into a diagonal matrix

    assert np.isclose(np.linalg.det(A), 1), "Determinant of A is not 1"

    # 5. Extracting K1
    K1 = U_prime @ np.conj(K2).T @ np.conj(A).T

    assert np.allclose(K1 @ K1.T, np.identity(4)), "K1 is not orthogonal"
    assert np.isclose(np.linalg.det(K1), 1), "Determinant of K1 is not 1"

    # 6. Extracting Local Gates
    L = B @ K1 @ np.conj(B).T  # Left Local Product
    R = B @ K2 @ np.conj(B).T  # Right Local Product

    phase1, L1, L2 = decompose_one_qubit_product(L)  # L1 (top), L2(bottom)
    phase2, R1, R2 = decompose_one_qubit_product(R)  # R1 (top), R2(bottom)

    # 7. Extracting the Canonical Parameters
    C = np.array([[1, 1, 1], [-1, 1, -1], [1, -1, -1]])  # Coefficient Matrix

    theta_vec = np.angle(np.diag(A))[:3]  # theta vector
    a0, a1, a2 = np.linalg.inv(C) @ theta_vec  # Computing the "a"-vector

    CAN = lambda c0, c1, c2: expm(1j / 2 * (c0 * np.kron(X, X) + c1 * np.kron(Y, Y) + c2 * np.kron(Z, Z)))
    # 8. Unpack Parameters and Put into Weyl chamber
    c0, c1, c2 = 2 * a1, -2 * a0, 2 * a2
    # print(c0, c1, c2)
    # print((phase1 * np.kron(L1, L2)) @ CAN(c0, c1, c2)
    #                   @ (phase2 * np.kron(R1, R2)))
    # c0, c1, c2 = weyl_chamber(np.array([2 * a1, -2 * a0, 2 * a2]))
    # print(c0, c1, c2)
    # print((phase1 * np.kron(L1, L2)) @ CAN(c0, c1, c2)
    #                   @ (phase2 * np.kron(R1, R2)))
    assert np.allclose(U, (phase1 * np.kron(L1, L2)) @ CAN(c0, c1, c2)
                       @ (phase2 * np.kron(R1, R2))), "U does not equal KAK"

    assert np.allclose(U, phase1 * phase2 * (np.kron(L1, L2)) @ CAN(c0, c1, c2)
                       @ (np.kron(R1, R2))), "U does not equal KAK"
    # print(np.allclose(U, phase1 *phase2 *( np.kron(L1, L2)) @ CAN(c0, c1, c2)
    #                   @ ( np.kron(R1, R2))))
    return phase1, L1, L2, phase2, R1, R2, c0, c1, c2

# U = unitary_group.rvs(4, random_state = 1)
# print(U)
# print(KAK_decomposition(U))
# print(np.kron(Y,Y))



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the Transpiler class"""

import time

from qsteed.passes.model import Model
from qsteed.passflow.passflow import PassFlow
from qsteed.transpiler.transpiler import Transpiler
from tests.shared_utils import get_passflow, get_random_circuit, get_initial_model


class TestTranspiler:
    """Test cases for the Transpiler class and related functionalities."""

    def test_random_circuit(self):
        """Test the generation of a random quantum circuit."""
        qc = get_random_circuit()
        assert qc is not None
        # qc.plot_circuit()
        # plt.show()
        qc.draw_circuit()

    def test_backend_model(self):
        """Test the generation of the initial backend model."""
        model = get_initial_model()
        assert model is not None
        assert isinstance(model, Model)

    def test_passflow(self):
        """Test the generation of PassFlow."""
        passflow = get_passflow()
        assert passflow is not None
        assert isinstance(passflow, PassFlow)

    def test_transpiler_custom(self):
        """
        Test the custom Transpiler process.
        Compile the quantum circuit using a custom passflow and initial model.
        """
        qc = get_random_circuit()
        st = time.time()
        passflow = get_passflow()
        initial_model = get_initial_model()
        transpiler = Transpiler(passflow, initial_model)
        transpiled_circuit = transpiler.transpile(qc)
        assert transpiled_circuit is not None
        print('qsteed time (s):', time.time() - st)
        print('qsteed depth:', len(transpiled_circuit.layered_circuit().T) - 1)
        print('qsteed swap count', transpiler.model.datadict['add_swap_count'])
        transpiled_circuit.draw_circuit()
        # transpiled_circuit.plot_circuit()
        # plt.show()

    def test_transpiler_preset(self):
        """
        Test the preset optimization level Transpiler process.
        Compile the quantum circuit using a preset optimization level.
        """
        qc = get_random_circuit()
        initial_model = get_initial_model()
        transpiler = Transpiler(initial_model=initial_model)
        transpiled_circuit = transpiler.transpile(qc, optimization_level=1)
        assert transpiled_circuit is not None
        transpiled_circuit.draw_circuit()
        # transpiled_circuit.plot_circuit()
        # plt.show()



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import pi
from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import WGate, RYGate, RZGate

from qsteed.passes.basepass import UnrollPass


class WToRYRZ(UnrollPass):
    """The WToRYRZ pass.
    convert W gate to {RY, RZ}.

    W gate decomposition rule:
    q[0] ----W---- ≡ U3(π,-π/4,π/4) * global_phase(π/2)
                   ≡ q[0] ----RZ(π/2)---RY(π)---- * global_phase(π/2)
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = WGate.name.lower()
        self.basis = [RYGate(0, 0).name.lower(), RZGate(0, 0).name.lower()]
        self.global_phase = pi / 2
        # qc = QuantumCircuit(1)
        # qc.rz(0, pi / 2)
        # qc.ry(0, pi)
        # self.circuit = qc

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, WGate):
            if isinstance(op.pos, list):
                pos = op.pos[0]
            else:
                pos = op.pos
            rule.append(RZGate(pos, pi / 2))
            rule.append(RYGate(pos, pi))
        else:
            rule.append(op)
        self.rule = rule
        return rule



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from math import pi
from typing import List

from quafu import QuantumCircuit
from quafu.elements import Barrier, Delay, XYResonance, Measure
from quafu.elements.element_gates import CXGate, RXGate, RYGate, RZGate, IdGate, CPGate, CZGate, ISwapGate
from quafu.exceptions.quafu_error import CompileError

from qsteed.passes.basepass import BasePass
from qsteed.passes.unroll.rules_library import Rules_dict, CX_rules


class UnrollToBasis(BasePass):
    """Unroll a circuit to a given basis gates.
    """

    def __init__(self, basis_gates: List[str] = None):
        """Initialize UnrollToBasis

        Args:
            basis_gates (list): e.g. ['cx','rx','ry','rz','id']
        """
        super().__init__()
        self.quantum_element = [Barrier.name, Delay.name, XYResonance.name, Measure.name]
        if basis_gates is None:
            self.basis_gates = [CXGate.name, RXGate.name, RYGate.name, RZGate.name, IdGate.name]
            # self.basis_gates = ['CX', 'RX', 'RY', 'RZ', 'Id']
        else:
            self.basis_gates = basis_gates
        self.basis_gates = self.basis_gates + self.quantum_element
        self.basis_gates = [gate.lower() for gate in self.basis_gates]
        self.gate_run_limit = 8
        self.qubits = None
        self.global_phase = 0

        if CZGate.name.lower() in self.basis_gates:
            Rules_dict.update(CX_rules[CZGate.name.lower()])
        elif ISwapGate.name.lower() in self.basis_gates:
            Rules_dict.update(CX_rules[ISwapGate.name.lower()])
        elif CPGate(0, 1, 0).name.lower() in self.basis_gates:
            Rules_dict.update(CX_rules[CPGate(0, 1, 0).name.lower()])

    def run(self, circuit: QuantumCircuit) -> QuantumCircuit:
        gates = circuit.gates
        self.qubits = circuit.num
        new_circuit = QuantumCircuit(circuit.num)
        for op in gates:
            current_op = copy.deepcopy(op)
            new_circuit = self._apply_gate_rules(current_op, op, new_circuit=new_circuit, depth=self.gate_run_limit)
        [new_circuit.measure([q], [c]) for q, c in circuit.measures.items()]
        self.global_phase = self.global_phase % (2 * pi)
        return new_circuit

    def _apply_gate_rules(self, current_op, gate, new_circuit=None, depth=5):
        if new_circuit is None:
            new_circuit = QuantumCircuit(self.qubits)

        if depth <= 0:
            # Exceeding the number of recursions indicates that no decomposition of this gate
            # based on the given basis gates set has been found among all rules.
            raise CompileError("The instruction %s cannot be unrolled to the given basis gates set %s." % (
                current_op.name.lower(), self.basis_gates))

        if gate.name.lower() in self.basis_gates:
            new_circuit.add_gate(gate)
            return new_circuit
        elif gate.name.lower() == 'unitary':
            # TODO: Currently quafu has no definition of unitary gate.
            pass
        else:
            try:
                Rules_dict[gate.name.lower()][0]
            except KeyError:
                print("Error: The instruction %s has no unrolling rule." % gate.name)
            basis = Rules_dict[gate.name.lower()][0]
            rule_class = Rules_dict[gate.name.lower()][1]
            rule = rule_class.run(gate)
            self.global_phase += rule_class.global_phase
            if set(basis).issubset(set(self.basis_gates)):
                [new_circuit.add_gate(g) for g in rule]
                return new_circuit
            else:
                for gate in rule:
                    new_circuit = self._apply_gate_rules(current_op, gate, new_circuit, depth=depth - 1)
                return new_circuit



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import ToffoliGate, CXGate, HGate, TGate, TdgGate

from qsteed.passes.basepass import UnrollPass


class ToffoliToCNOT(UnrollPass):
    """The ToffoliToCNOT pass.
    convert Toffoli(CCX) gate to {CNOT, H, Tdg, T}.

    Toffoli(CCX) gate decomposition rule:
    q[0]  --*--      q[0]  -------------------*---------------------*----*-----T-----*--
            |                                 |                     |    |           |
    q[1]  --*--  ≡   q[1]  -------*-----------|---------*-----T-----|----+----Tdg----+--
            |                     |           |         |           |
    q[2]  --X--      q[2]  --H----+----Tdg----+----T----+----Tdg----+----T-----H--------
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = ToffoliGate(0, 1, 2).name.lower()
        self.basis = [CXGate.name.lower(), HGate.name.lower(), TGate.name.lower(), TdgGate.name.lower()]

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, ToffoliGate):
            rule.append(HGate(op.pos[2]))
            rule.append(CXGate(op.pos[1], op.pos[2]))
            rule.append(TdgGate(op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[2]))
            rule.append(TGate(op.pos[2]))
            rule.append(CXGate(op.pos[1], op.pos[2]))
            rule.append(TGate(op.pos[1]))
            rule.append(TdgGate(op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(TGate(op.pos[2]))
            rule.append(TGate(op.pos[0]))
            rule.append(TdgGate(op.pos[1]))
            rule.append(HGate(op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
        else:
            rule.append(op)
        self.rule = rule
        return rule


class ToffoliToCNOT8(UnrollPass):
    """The ToffoliToCNOT pass.
    convert Toffoli(CCX) gate to {CNOT, H, Tdg, T}.

    Toffoli(CCX) gate decomposition rule:
    q[0]  --*--      q[0]  -------T----*---------*---------*----------------*--------------
            |                          |         |         |                |           
    q[1]  --*--  ≡   q[1]  -------T----+----*----+----*----+----Tdg----*----+------*-------
            |                               |         |                |           |
    q[2]  --X--      q[2]  --H----T---------+----T----+---------Tdg----+----Tdg----+----H--
    """

    def __init__(self) -> None:
        super().__init__()
        self.original = ToffoliGate(0, 1, 2).name.lower()
        self.basis = [CXGate.name.lower(), HGate.name.lower(), TGate.name.lower(), TdgGate.name.lower()]

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, ToffoliGate):
            rule.append(HGate(op.pos[2]))
            rule.append(TGate(op.pos[0]))
            rule.append(TGate(op.pos[1]))
            rule.append(TGate(op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(CXGate(op.pos[1], op.pos[2]))
            rule.append(TGate(op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(CXGate(op.pos[1], op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(TdgGate(op.pos[1]))
            rule.append(TdgGate(op.pos[2]))
            rule.append(CXGate(op.pos[1], op.pos[2]))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(TdgGate(op.pos[2]))
            rule.append(CXGate(op.pos[1], op.pos[2]))
            rule.append(HGate(op.pos[2]))
        else:
            rule.append(op)
        self.rule = rule
        return rule



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from typing import Dict, Any, List, Union


@dataclasses.dataclass
class InstructionNode:
    """
    A class representing a single instruction in a quantum circuit.

    Attributes:
    -----------
    name : Any
        The name of the gate.
    pos : Union[List[Any], Dict[Any, Any]]
        The position of the gate in the circuit. If the gate is a measurement gate, it is a dictionary with qubit indices as keys and classical bit indices as values.
    paras : List[Any]
        The parameters of the gate.
    duration : int
        The duration of the gate. Only applicable for certain gates.
    unit : str
        The unit of the duration. Only applicable for certain gates.
    label : Union[str, int]
        The label of the instruction node.

    Methods:
    --------
    __hash__(self)
        Returns the hash value of the instruction node.
    __str__(self)
        Returns a string representation of the instruction node.
    __repr__(self)
        Returns a string representation of the instruction node.
    """
    # def __init__(self, name: Any, pos: Union[List[Any], Dict[Any, Any]], paras: List[Any], duration: int, unit: str,
    #              label: Union[str, int]):
    #     self.name = name
    #     self.pos = pos
    #     self.paras = paras
    #     self.duration = duration
    #     self.unit = unit
    #     self.label = label

    name: Any  # gate.name
    pos: Union[List[Any], Dict[Any, Any]]  # gate.pos |  Dict[Any,Any] for measure
    paras: List[Any]  # gate.paras
    # matrix:List[Any]   # for gate in [QuantumGate]
    duration: int  # for gate in [Delay,XYResonance] in quafu
    unit: str  # for gate in [Delay,XYResonance] in quafu
    label: Union[str, int]

    def __hash__(self):
        return hash((type(self.name), tuple(self.pos), self.label))

    def __str__(self):
        if self.name == 'measure':
            args = ','.join(str(q) for q in self.pos.keys())
            args += f'=>{",".join(str(c) for c in self.pos.values())}'
        else:
            args = ','.join(str(q) for q in self.pos)

        if self.paras is None:
            return f'{self.label}{{{self.name}({args})}}'
        else:
            # if self.paras not a list, then make it a list  of str of .3f float
            if not isinstance(self.paras, list):
                if isinstance(self.paras, float):
                    formatted_paras = [f'{self.paras:.3f}']
                else:
                    formatted_paras = [str(self.paras)]
            else:
                if all(isinstance(p, float) for p in self.paras):
                    formatted_paras = [f'{p:.3f}' for p in self.paras]
                else:
                    formatted_paras = [str(p) for p in self.paras]

            formatted_paras_str = ','.join(formatted_paras)
            return f'{self.label}{{{self.name}({args})}}({formatted_paras_str})'

    def __repr__(self):
        return str(self)



# This code is part of QSteed.
#
# (C) Copyright 2024 Beijing Academy of Quantum Information Sciences
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from quafu.elements import Instruction
from quafu.elements.element_gates import CPGate, CXGate, PhaseGate

from qsteed.passes.basepass import UnrollPass


class CPToCNOT(UnrollPass):
    """The CPToCNOT pass.

    convert CPhase gate to {CNOT, Phase}.

    CP gate decomposition rule:
    q[0]  -----*------     q[0]  --P(theta/2)----*-------------------*----------------
               |        ≡                        |                   |
    q[1]  --P(theta)--     q[1]  ----------------+----P(-theta/2)----+----P(theta/2)--
    """

    def __init__(self) -> None:
        super().__init__()
        self.parameter_type = 'parameterized_gate'
        self.original = CPGate(0, 1, 0).name.lower()
        self.basis = [CXGate.name.lower(), PhaseGate(0, 0).name.lower()]
        # qc = QuantumCircuit(2)
        # qc.p(0, theta/2)
        # qc.cnot(0, 1)
        # qc.p(1, -theta/2)
        # qc.cnot(0, 1)
        # qc.p(1, theta/2)
        # self.circuit = qc

    def run(self, op: Instruction) -> List[Instruction]:
        rule = []
        if isinstance(op, CPGate):
            if isinstance(op.paras, list):
                theta = op.paras[0]
            else:
                theta = op.paras
            rule.append(PhaseGate(op.pos[0], theta / 2))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(PhaseGate(op.pos[1], -theta / 2))
            rule.append(CXGate(op.pos[0], op.pos[1]))
            rule.append(PhaseGate(op.pos[1], theta / 2))
        else:
            rule.append(op)
        self.rule = rule
        return rule

    # def run(self, circuit: QuantumCircuit) -> QuantumCircuit:
    #     new_circuit = QuantumCircuit(circuit.num)
    #     for op in circuit.gates:
    #         if isinstance(op, CPGate):
    #             new_circuit.add_gate(PhaseGate(op.pos[0],theta/2))
    #             new_circuit.add_gate(CXGate(op.pos[0], op.pos[1]))
    #             new_circuit.add_gate(PhaseGate(op.pos[1],-theta/2))
    #             new_circuit.add_gate(CXGate(op.pos[0],op.pos[1]))
    #             new_circuit.add_gate(PhaseGate(op.pos[1],theta/2))
    #         else:
    #             new_circuit.add_gate(op)
    #     return new_circuit
